/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QBrush>
#include <QColor>
#include <QGridLayout>
#include <QLabel>
#include <QFrame>
#include <QPlainTextEdit>
#include <QLineEdit>
#include <QPalette>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QFont>

#include "activeprogramwidget.h"

TActiveProgramWidget::TActiveProgramWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QLabel								*collecting_data_label;
	QLabel								*program_name_label;
	QLabel								*program_revision_label;
	QLabel								*execution_date_label;
	QLabel								*execution_time_label;
	QLabel								*execution_time_units_label;
	QLabel								*idle_time_label;
	QLabel								*idle_time_units_label;
	QLabel								*open_time_label;
	QLabel								*open_time_units_label;
	QLabel								*last_error_label;
	QLabel								*point_count_label;
	QSpacerItem							*control_hspacer;
	QFont								text_font;
	
	text_font.setFamily(QStringLiteral("Courier New"));
	text_font.setFixedPitch(true);

	widget_layout = new QGridLayout(this);
	
	collecting_data_label = new QLabel(this);
	widget_layout->addWidget(collecting_data_label,0,0,1,1);

	d_collecting_data_frame = new QFrame(this);
	d_collecting_data_frame->setFrameShape(QFrame::StyledPanel);
	d_collecting_data_frame->setFrameShadow(QFrame::Plain);
	widget_layout->addWidget(d_collecting_data_frame,0,1,1,3);

	program_name_label = new QLabel(this);
	widget_layout->addWidget(program_name_label,1,0,1,1);

	d_program_name_edit = new QLineEdit(this);
	d_program_name_edit->setReadOnly(true);
	widget_layout->addWidget(d_program_name_edit,1,1,1,3);

	program_revision_label = new QLabel(this);
	widget_layout->addWidget(program_revision_label,2,0,1,1);

	d_program_revision_edit = new QLineEdit(this);
	d_program_revision_edit->setReadOnly(true);
	widget_layout->addWidget(d_program_revision_edit,2,1,1,3);

	execution_date_label = new QLabel(this);
	widget_layout->addWidget(execution_date_label,3,0,1,1);

	d_execution_date_edit = new QLineEdit(this);
	d_execution_date_edit->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Fixed);
	d_execution_date_edit->setReadOnly(true);
	widget_layout->addWidget(d_execution_date_edit,3,1,1,1);

	control_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	widget_layout->addItem(control_hspacer,4,3,3,1);

	open_time_label = new QLabel(this);
	widget_layout->addWidget(open_time_label,4,0,1,1);

	d_open_time_edit = new QLineEdit(this);
	d_open_time_edit->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Fixed);
	d_open_time_edit->setAlignment(Qt::AlignCenter);
	d_open_time_edit->setReadOnly(true);
	widget_layout->addWidget(d_open_time_edit,4,1,1,1);

	open_time_units_label = new QLabel(this);
	widget_layout->addWidget(open_time_units_label,4,2,1,1);

	execution_time_label = new QLabel(this);
	widget_layout->addWidget(execution_time_label,5,0,1,1);

	d_execution_time_edit = new QLineEdit(this);
	d_execution_time_edit->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Fixed);
	d_execution_time_edit->setAlignment(Qt::AlignCenter);
	d_execution_time_edit->setReadOnly(true);
	widget_layout->addWidget(d_execution_time_edit,5,1,1,1);

	execution_time_units_label = new QLabel(this);
	widget_layout->addWidget(execution_time_units_label,5,2,1,1);

	idle_time_label = new QLabel(this);
	widget_layout->addWidget(idle_time_label,6,0,1,1);

	d_idle_time_edit = new QLineEdit(this);
	d_idle_time_edit->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Fixed);
	d_idle_time_edit->setAlignment(Qt::AlignCenter);
	d_idle_time_edit->setReadOnly(true);
	widget_layout->addWidget(d_idle_time_edit,6,1,1,1);

	idle_time_units_label = new QLabel(this);
	widget_layout->addWidget(idle_time_units_label,6,2,1,1);

	point_count_label = new QLabel(this);
	widget_layout->addWidget(point_count_label,7,0,1,1);
	
	d_point_count_edit = new QLineEdit(this);
	d_point_count_edit->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Fixed);
	d_point_count_edit->setAlignment(Qt::AlignCenter);
	d_point_count_edit->setReadOnly(true);
	widget_layout->addWidget(d_point_count_edit,7,1,1,1);

	last_error_label = new QLabel(this);
	widget_layout->addWidget(last_error_label,8,0,1,1);

	d_last_error_edit = new QLineEdit(this);
	d_last_error_edit->setReadOnly(true);
	widget_layout->addWidget(d_last_error_edit,8,1,1,3);
	
	d_message_text = new QPlainTextEdit();
	d_message_text->setFont(text_font);
	d_message_text->setWordWrapMode(QTextOption::NoWrap);
	d_message_text->setReadOnly(true);
	widget_layout->addWidget(d_message_text,9,0,1,4);
	
	// defaults
	d_collecting_data_frame->setMinimumHeight(d_program_name_edit->sizeHint().height());
	
	QWidget::setTabOrder(d_program_name_edit,d_program_revision_edit);
	QWidget::setTabOrder(d_program_revision_edit,d_execution_date_edit);
	QWidget::setTabOrder(d_execution_date_edit,d_open_time_edit);
	QWidget::setTabOrder(d_open_time_edit,d_execution_time_edit);
	QWidget::setTabOrder(d_execution_time_edit,d_idle_time_edit);
	QWidget::setTabOrder(d_idle_time_edit,d_point_count_edit);
	QWidget::setTabOrder(d_point_count_edit,d_last_error_edit);
	QWidget::setTabOrder(d_last_error_edit,d_message_text);

	collecting_data_label->setText(QStringLiteral("Collection Active:"));
	program_name_label->setText(QStringLiteral("Program Name:"));
	program_revision_label->setText(QStringLiteral("Program Revision:"));
	execution_date_label->setText(QStringLiteral("Execution Date:"));
	open_time_label->setText(QStringLiteral("Open Time:"));
	idle_time_units_label->setText(QStringLiteral("sec"));
	last_error_label->setText(QStringLiteral("Last Error Message:"));
	idle_time_label->setText(QStringLiteral("Idle Time:"));
	execution_time_label->setText(QStringLiteral("Execution Time:"));
	point_count_label->setText(QStringLiteral("Point Count:"));
	execution_time_units_label->setText(QStringLiteral("sec"));
	open_time_units_label->setText(QStringLiteral("sec"));
	
	this->Reset();

}

TActiveProgramWidget::~TActiveProgramWidget(void)
{
}

int TActiveProgramWidget::Open_Time(void) const
{
	return d_open_time_edit->text().toInt();
}

int TActiveProgramWidget::Idle_Time(void) const
{
	return d_idle_time_edit->text().toInt();
}

int TActiveProgramWidget::Execution_Time(void) const
{
	return d_execution_time_edit->text().toInt();
}

void TActiveProgramWidget::Enable_Collection(
	const bool 							state)
{
	if(state)
	{
		d_collecting_data_frame->setStyleSheet(QString::fromUtf8("background-color: rgb(0, 128, 0);"));
	}
	else
	{
		d_collecting_data_frame->setStyleSheet(QString());
	}
}

void TActiveProgramWidget::Set_Name(
	const QString 						&text)
{
	d_program_name_edit->setText(text);
}

void TActiveProgramWidget::Set_Revision(
	const QString 						&text)
{
	d_program_revision_edit->setText(text);
}

void TActiveProgramWidget::Set_Date(
	const QString					 	&text)
{
	d_execution_date_edit->setText(text);
}

void TActiveProgramWidget::Increment_Open_Time_MS(
	const int 							milliseconds)
{
	int									idle_time;
	
	d_open_time_milliseconds += milliseconds;
	d_open_time_edit->setText(QString("%1").arg(d_open_time_milliseconds / 1000));
	
	idle_time = static_cast<int>(static_cast<double>((d_open_time_milliseconds - d_execution_time_milliseconds) / 1000) + 0.5);
	d_idle_time_edit->setText(QString("%1").arg(idle_time));
}

void TActiveProgramWidget::Increment_Execution_Time_MS(
	const int 							milliseconds)
{
	int									idle_time;

	d_execution_time_milliseconds += milliseconds;
	d_execution_time_edit->setText(QString("%1").arg(d_execution_time_milliseconds / 1000));
	
	idle_time = static_cast<int>(static_cast<double>((d_open_time_milliseconds - d_execution_time_milliseconds) / 1000) + 0.5);
	d_idle_time_edit->setText(QString("%1").arg(idle_time));
}

void TActiveProgramWidget::Set_Point_Count(
	const int 							count)
{
	d_point_count_edit->setText(QString("%1").arg(count));
}

void TActiveProgramWidget::Set_Last_Error(
	const QString 						&text)
{
	d_last_error_edit->setText(text);
}

void TActiveProgramWidget::Reset(void)
{
	d_program_name_edit->setText(QString());
	d_program_revision_edit->setText(QString());
	d_execution_date_edit->setText(QString());
	d_open_time_edit->setText(QStringLiteral("0"));
	d_idle_time_edit->setText(QStringLiteral("0"));
	d_execution_time_edit->setText(QStringLiteral("0"));
	d_point_count_edit->setText(QStringLiteral("0"));
	d_last_error_edit->setText(QString());
	
	d_execution_time_milliseconds = 0;
	d_open_time_milliseconds = 0;
	
	this->Clear_Messages();
}

void TActiveProgramWidget::Add_Message(
const QString						&text)
{
	d_message_text->appendPlainText(text);
}

void TActiveProgramWidget::Clear_Messages(void)
{
	d_message_text->clear();
}


