/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QScreen>
#include <QAction>
#include <QToolBar>
#include <QStatusBar>
#include <QLabel>
#include <QWidget>
#include <QPlainTextEdit>
#include <QTabWidget>
#include <QMenu>
#include <QMenuBar>
#include <QFrame>
#include <QTimer>
#include <QDateTime>
#include <QThread>
#include <QIcon>
#include <QStringListModel>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QSizePolicy>
#include <QCloseEvent>
#include <QSettings>
#include <QFileDialog>
#include <QFile>
#include <QDir>
#include <QRect>
#include <QSize>
#include <QFont>
#include <cmath>
#include <assert.h>

#include "../../core/messagebox.h"

#include "monitorprocess.h"
#include "datawidget.h"
#include "database.h"
#include "activeprogramwidget.h"
#include "summarytext.h"

#include "cmmusage.h"

TCMMUsage::TCMMUsage(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QAction								*action_help;
	QAction								*action_quit;
	QToolBar							*toolbar;
	QWidget								*central_widget;
	QLabel								*copyright_text_label;
	QLabel								*copyright_logo_label;
	QGridLayout							*widget_layout;
	QHBoxLayout							*copyright_hlayout;
	QFrame								*separator_line;
	QMenu								*file_menu;
	QMenu								*help_menu;
	QMenuBar							*menu_bar;
	QRect								desktop_rect;
	QRect								window_rect;
	QPoint								position;
	QSize								size;
	QDir								settings_dir;
	QFont								copyright_font;
	QFont								text_font;

	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	text_font.setFamily(QStringLiteral("Courier New"));
	text_font.setFixedPitch(true);

	settings_dir = QDir::home();
	
	if(!settings_dir.exists(".cmmusage"))
	{
		settings_dir.mkdir(".cmmusage");
	}
	
	settings_dir.cd(".cmmusage");
	
	d_data_path = settings_dir.absolutePath();
		
	d_settings = new QSettings(d_data_path + QStringLiteral("/Settings.ini"),QSettings::IniFormat,this);
	
	position = d_settings->value(QStringLiteral("Mainwindow_Position"), QPoint(30,30)).toPoint();
	size = d_settings->value(QStringLiteral("Mainwindow_Size"), QSize(801,601)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid() && window_rect.width() > 10 && window_rect.height() > 10)
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(801,601);
		this->move(QPoint(30,30));
	}
	
	d_export_path = d_settings->value(QStringLiteral("Export_Path"), QDir::homePath()).toString();
	
	action_quit = new QAction(this);
	action_help = new QAction(this);

	d_connect_disconnect_action = new QAction(this);
	d_connect_disconnect_action->setIcon(QIcon(":/toolbar/not_connected_x32.png"));
	
	d_export_all_csv_action = new QAction(this);
	d_export_all_csv_action->setIcon(QIcon(":/toolbar/export_all_csv_x32.png"));
	
	d_export_selected_csv_action = new QAction(this);
	d_export_selected_csv_action->setIcon(QIcon(":/toolbar/export_selected_csv_x32.png"));
	
	d_export_selected_txt_action = new QAction(this);
	d_export_selected_txt_action->setIcon(QIcon(":/toolbar/export_selected_txt_x32.png"));

	menu_bar = new QMenuBar(this);
	this->setMenuBar(menu_bar);

	file_menu = new QMenu(menu_bar);
	menu_bar->addMenu(file_menu);
	file_menu->addAction(d_connect_disconnect_action);
	file_menu->addSeparator();
	file_menu->addAction(d_export_all_csv_action);
	file_menu->addAction(d_export_selected_csv_action);
	file_menu->addAction(d_export_selected_txt_action);
	file_menu->addSeparator();
	file_menu->addAction(action_quit);

	help_menu = new QMenu(menu_bar);
	menu_bar->addMenu(help_menu);
	help_menu->addAction(action_help);

	menu_bar->addAction(file_menu->menuAction());
	menu_bar->addAction(help_menu->menuAction());

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);
	
	widget_layout = new QGridLayout(central_widget);

	d_date_data_widget = new TDataWidget(central_widget);
	d_date_data_widget->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Expanding);
 	widget_layout->addWidget(d_date_data_widget,0,0,1,1);

	d_program_data_widget = new TDataWidget(central_widget);
	d_program_data_widget->setSizePolicy(QSizePolicy::Maximum,QSizePolicy::Expanding);
 	widget_layout->addWidget(d_program_data_widget,1,0,1,1);
	
	d_view_tab = new QTabWidget(central_widget);
	
	d_active_program_widget = new TActiveProgramWidget();
	d_view_tab->addTab(d_active_program_widget,QStringLiteral("Active"));
	
	d_summary_text = new QPlainTextEdit();
	d_summary_text->setFont(text_font);
	d_summary_text->setWordWrapMode(QTextOption::NoWrap);
	d_summary_text->setSizePolicy(QSizePolicy::MinimumExpanding,QSizePolicy::Expanding);
	d_summary_text->setReadOnly(true);
	d_summary_text->setUndoRedoEnabled(false);
	d_view_tab->addTab(d_summary_text,QStringLiteral("Selected"));

 	widget_layout->addWidget(d_view_tab,0,1,2,1);

	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	this->addToolBar(Qt::TopToolBarArea,toolbar);
	
	toolbar->addAction(d_connect_disconnect_action);
	toolbar->addSeparator();
	toolbar->addAction(d_export_all_csv_action);
	toolbar->addAction(d_export_selected_csv_action);
	toolbar->addSeparator();
	toolbar->addAction(d_export_selected_txt_action);

	d_status_bar = new QStatusBar(this);
	this->setStatusBar(d_status_bar);
	
	separator_line = new QFrame(this);
	separator_line->setFrameShape(QFrame::HLine);
	separator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_line,2,0,1,2);
	
	copyright_hlayout = new QHBoxLayout();

	copyright_text_label = new QLabel(central_widget);
	copyright_text_label->setFont(copyright_font);
	copyright_text_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_text_label);
	
	copyright_logo_label = new QLabel(central_widget);
	copyright_logo_label->setMinimumSize(100,40);
	copyright_logo_label->setMaximumSize(100,40);
	copyright_hlayout->addWidget(copyright_logo_label);
	widget_layout->addLayout(copyright_hlayout,3,0,1,2);
	
#ifndef Q_OS_WIN
	d_simulation_timer = new QTimer(this);
	d_simulation_timer->setInterval(750);
#endif

	// defaults
	d_pcdmis_application_events = 0;
	d_pcdmis_machine_events = 0;
	d_summary_thread = 0;
	d_summary_text_worker = 0;
	d_database = new TDatabase;
	d_update_timer = new QTimer(this);
	d_monitor_process = new TMonitorProcess(this);
	d_msg_box = new TMessageBox(this);
	d_connection_status = TCMMUsage::NOT_CONNECTED;
	d_monitor_active = false;
	d_program_open = false;
	d_program_executing = false;
	d_export_selected_csv_action->setEnabled(false);
	d_export_selected_txt_action->setEnabled(false);

	if(this->Open_Database(d_data_path + QStringLiteral("/database.db")))
	{
		d_date_data_widget->Set_Data(d_database->Dates());
		d_program_data_widget->Set_Data(d_database->Programs(QStringList()));
	}
	
	this->Set_Status_Text(QStringLiteral("INF:  Not Connected."));
	copyright_logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));

	this->setWindowTitle(QStringLiteral("CMM Usage"));
	
	action_quit->setText(QStringLiteral("Quit"));
	action_help->setText(QStringLiteral("About"));
	file_menu->setTitle(QStringLiteral("File"));
	help_menu->setTitle(QStringLiteral("Help"));
	d_connect_disconnect_action->setText(QStringLiteral("Monitor"));
	d_export_all_csv_action->setText(QStringLiteral("Export All CSV"));
	d_export_selected_csv_action->setText(QStringLiteral("Export CSV"));
	d_export_selected_txt_action->setText(QStringLiteral("Export Text"));
	copyright_text_label->setText(QStringLiteral("Copyright (C) 2024 Select Calibration Incorporated.  www.selectcalibration.ca"));
	d_date_data_widget->Set_Title(QStringLiteral("Date"));
	d_program_data_widget->Set_Title(QStringLiteral("Program"));
	d_summary_text->setPlainText(QStringLiteral("No summary available."));
	
	d_connect_disconnect_action->setToolTip(QStringLiteral("Enable or disable monitoring events."));
	d_export_all_csv_action->setToolTip(QStringLiteral("Export all data to a CSV file."));
	d_export_selected_csv_action->setToolTip(QStringLiteral("Export selected data to a CSV file."));
	d_export_selected_txt_action->setToolTip(QStringLiteral("Export selected data to a TXT file."));
	
	connect(d_connect_disconnect_action,&QAction::triggered,this,&TCMMUsage::Toggle_Monitor);
	connect(d_export_all_csv_action,&QAction::triggered,this,&TCMMUsage::Export_All_CSV);
	connect(d_export_selected_csv_action,&QAction::triggered,this,&TCMMUsage::Export_Selected_CSV);
	connect(d_export_selected_txt_action,&QAction::triggered,this,&TCMMUsage::Export_Selected_TXT);
	connect(action_quit,&QAction::triggered,this,&TCMMUsage::close);
	connect(action_help,&QAction::triggered,this,&TCMMUsage::Help_About);

	connect(d_monitor_process,&TMonitorProcess::started,this,&TCMMUsage::Monitor_Started);
	connect(d_monitor_process,&TMonitorProcess::finished,this,&TCMMUsage::Monitor_Finished);
	connect(d_monitor_process,&TMonitorProcess::Class_Visible,this,&TCMMUsage::Suspend_Resume);
	
#ifndef Q_OS_WIN	
	d_simulation_events.push_back("OnOpenPartProgram");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnStartExecution");
	d_simulation_events.push_back("OnObjectExecuted");
	d_simulation_events.push_back("OnObjectAboutToExecute");
	d_simulation_events.push_back("OnEndExecution");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnStartExecution");
	d_simulation_events.push_back("OnObjectExecuted");
	d_simulation_events.push_back("OnObjectAboutToExecute");
	d_simulation_events.push_back("OnEndExecution");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnUpdateStatusMessage");
	d_simulation_events.push_back("OnClosePartProgram");

	connect(d_simulation_timer,&QTimer::timeout,this,&TCMMUsage::Simulation_Timeout);
#endif
	
	d_update_timer->setInterval(1000);

	connect(d_date_data_widget,&TDataWidget::Selection_Changed,this,&TCMMUsage::Date_Selection_Changed);
	connect(d_program_data_widget,&TDataWidget::Selection_Changed,this,&TCMMUsage::Program_Selection_Changed);
	connect(d_update_timer,&QTimer::timeout,this,&TCMMUsage::Update_Timeout);
}

TCMMUsage::~TCMMUsage(void)
{
	delete d_database;
	d_database = 0;
}

void TCMMUsage::Toggle_Monitor(void)
{
	if(d_connection_status == TCMMUsage::NOT_CONNECTED)
	{
		d_connect_disconnect_action->setIcon(QIcon(":/toolbar/connecting_x32.png"));
		
		d_connection_status = TCMMUsage::CONNECTION_IDLE;
		this->Set_Status_Text(QStringLiteral("INF:  Connecting..."));
				
		d_monitor_process->Set_Monitor_Class(QStringLiteral("WilcoxAssociatesIncPC-DMISWndClass"));
		d_monitor_process->start(QThread::NormalPriority);
	}
	else
	{
		d_connection_status = TCMMUsage::NOT_CONNECTED;
		d_connect_disconnect_action->setIcon(QIcon(":/toolbar/not_connected_x32.png"));
		

#ifndef Q_OS_WIN
		d_simulation_timer->stop();
#endif
		
		d_program_open = false;
		d_program_executing = false;
		this->Update_Timeout();

		this->Write_Active_DatabaseItem();

		this->Disconnect_PCDMIS();
		
		d_monitor_process->Stop_Thread();
		
		this->Set_Status_Text(QStringLiteral("INF:  Disconnected."));
	}
}

void TCMMUsage::Suspend_Resume(
	bool								state)
{
	int									seconds;
		
	if(state && d_connection_status == TCMMUsage::CONNECTION_IDLE)
	{
		if(this->Connect_PCDMIS())
		{
			d_connect_disconnect_action->setIcon(QIcon(":/toolbar/connected_x32.png"));
			d_connection_status = TCMMUsage::CONNECTED;
			
			this->Set_Status_Text(QStringLiteral("INF:  Connected"));
			
			d_software_run_timer.start();
			d_update_timer->start();

#ifndef Q_OS_WIN
			d_simulation_timer->start();
#endif
			
		}
		else
		{
			d_connect_disconnect_action->setIcon(QIcon(":/toolbar/err_connect_x32.png"));
			d_connection_status = TCMMUsage::CONNECTION_IDLE;
			
			this->Set_Status_Text(QStringLiteral("INF:  Connection Error"));
		}
	}
	else
	{		
		if(d_connection_status == TCMMUsage::CONNECTED)
		{
			d_update_timer->stop();

			this->Disconnect_PCDMIS();
			
			seconds = d_software_run_timer.restart() / 1000;

			if(seconds > 0)
			{
				d_database->Increment_Software_Run_Time(seconds);
			}
		}
		
		seconds = d_monitor_timer.restart() / 1000;

		if(seconds > 0)
		{
			d_database->Increment_Software_Monitor_Time(seconds);
		}

		d_connection_status = TCMMUsage::CONNECTION_IDLE;
		d_connect_disconnect_action->setIcon(QIcon(":/toolbar/connecting_x32.png"));

		this->Set_Status_Text(QStringLiteral("INF:  Waiting for PC-DMIS software..."));
	}
}

void TCMMUsage::Application_Event(
	const QString						&event_name,
	int									argument_count,
	TAutomation::TArg					* const arguments)
{
	QString								text;
	QString								name;
	QString								revision;
	
	if(event_name == QStringLiteral("OnOpenPartProgram"))
	{
		d_active_database_item.Clear();
		d_program_executing = false;

		if(argument_count > 0)
		{
			TAutomation program((arguments + 0)->AsObjectInterface());
			
			if(!program.IsNullObject())
			{
				this->Open_PCDMIS_Program(&program);
				program.ReleaseObject();
			}
		}
		
#ifndef Q_OS_WIN
		this->Open_PCDMIS_Program(0);
#endif
	}
	else if(event_name == QStringLiteral("OnStartExecution"))
	{
		d_active_program_widget->Add_Message(QStringLiteral("INF:  Application Event: OnStartExecution"));

		d_program_execution_timer.start();
		d_program_executing = true;
	}
	else if(event_name == QStringLiteral("OnEndExecution"))
	{
		d_active_program_widget->Add_Message(QStringLiteral("INF:  Application Event: OnEndExecution"));
		
		d_program_executing = false;

		d_active_program_widget->Increment_Execution_Time_MS(d_program_execution_timer.restart());		
	}
	else if(event_name == QStringLiteral("OnClosePartProgram"))
	{
		if(argument_count > 0)
		{
			TAutomation program((arguments + 0)->AsObjectInterface());
			
			if(!program.IsNullObject())
			{
				this->Close_PCDMIS_Program(&program);
				program.ReleaseObject();
			}
		}
		
#ifndef Q_OS_WIN
		this->Close_PCDMIS_Program(0);
#endif
	}
}

void TCMMUsage::Machine_Event(
	const QString						&event_name,
	int									argument_count,
	TAutomation::TArg					* const arguments)
{
	QString								text;
	
	d_active_program_widget->Add_Message(QString("INF:  Machine Event:     %1").arg(event_name));
	
	if(event_name == QStringLiteral("LearnHit"))
	{
		if(argument_count == 2)
		{
			TAutomation program((arguments + 0)->AsObjectInterface());
			TAutomation command((arguments + 1)->AsObjectInterface());
			
			if(!command.IsNullObject())
			{
				command.ReleaseObject();
			}
			
			if(!program.IsNullObject())
			{
				program.ReleaseObject();
			}
		}
		
		d_active_database_item.Increment_Executed_Point_Count();
		d_active_program_widget->Set_Point_Count(d_active_database_item.Executed_Point_Count());
	}
	else if(event_name == QStringLiteral("ExecuteHit"))
	{
		d_active_database_item.Increment_Executed_Point_Count();
		d_active_program_widget->Set_Point_Count(d_active_database_item.Executed_Point_Count());
	}
	else if(event_name == QStringLiteral("ErrorMsg"))
	{
		text = (arguments + 0)->AsString().simplified();

		d_active_program_widget->Set_Last_Error(text);
		d_active_database_item.Add_Error_Message(text);
	}
}

void TCMMUsage::Export_All_CSV(void)
{
	QStringList							dates;
	QStringList							programs;
	QStringList::iterator				program_iter;
	QStringList::iterator				date_iter;
	std::vector<TDatabaseItem>			data_items_date;
	std::vector<TDatabaseItem>			data_items;
	std::vector<TDatabaseItem>::iterator data_iter;
	QString								file_name;
	QString								text;
	QFile								file;
	int									index;
	int									count;
	bool								init;
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Export All to CSV"),
											 d_export_path,
											 QStringLiteral("Comma Separated File (*.csv)"));
	
	if(file_name.length())
	{
		
		file.setFileName(file_name);
		
		if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
		{
			return;
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_export_path = file_name;
			d_export_path.truncate(index);
		}
		
		dates = d_database->Dates();
		programs = d_database->Programs(dates);
		
		for(program_iter = programs.begin();program_iter != programs.end();++program_iter)
		{
			for(date_iter = dates.begin();date_iter != dates.end();++date_iter)
			{
				data_items_date = d_database->Program_Data((*program_iter),(*date_iter));
				data_items.insert(data_items.end(),data_items_date.begin(),data_items_date.end());
			}
		}
		
		init = true;
		
		for(data_iter = data_items.begin();data_iter != data_items.end();++data_iter)
		{
			if(init)
			{
				file.write("Name,Revision,Date,Time,Execution Time,Idle Time,Point Count,Error Message Count,Error Messages\n");
				init = false;
			}
			
			count = (*data_iter).Error_Message_Count();
			
			text = QString("%1,%2,%3,%4,%5,%6,%7,%8")
						.arg((*data_iter).Name())
						.arg((*data_iter).Revision())
						.arg((*data_iter).Date())
						.arg((*data_iter).Time())
						.arg((*data_iter).Execution_Time())
						.arg((*data_iter).Idle_Time())
						.arg((*data_iter).Executed_Point_Count())
						.arg(count);
			
			for(index = 0;index < count;++index)
			{
				text += QString(",%1").arg((*data_iter).Error_Message(index));
			}
			
			text += '\n';
			
			file.write(text.toLatin1());
		
		}
		
		file.close();
		
	}
}

void TCMMUsage::Export_Selected_CSV(void)
{
	QStringList							dates;
	QStringList::iterator				program_iter;
	QStringList::iterator				date_iter;
	std::vector<TDatabaseItem>			data_items_date;
	std::vector<TDatabaseItem>			data_items;
	std::vector<TDatabaseItem>::iterator data_iter;
	QString								file_name;
	QString								text;
	QFile								file;
	int									index;
	int									count;
	bool								init;
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Export Selected to CSV"),
											 d_export_path,
											 QStringLiteral("Comma Separated File (*.csv)"));
	
	if(file_name.length())
	{
		
		file.setFileName(file_name);
		
		if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
		{
			return;
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_export_path = file_name;
			d_export_path.truncate(index);
		}
		
		if(d_selected_dates.size())
		{
			dates = d_selected_dates;
		}
		else
		{
			dates = d_database->Dates();
		}
				
		for(program_iter = d_selected_programs.begin();program_iter != d_selected_programs.end();++program_iter)
		{
			for(date_iter = dates.begin();date_iter != dates.end();++date_iter)
			{
				data_items_date = d_database->Program_Data((*program_iter),(*date_iter));
				data_items.insert(data_items.end(),data_items_date.begin(),data_items_date.end());
			}
		}
		
		init = true;
		
		for(data_iter = data_items.begin();data_iter != data_items.end();++data_iter)
		{
			if(init)
			{
				file.write("Name,Revision,Date,Time,Execution Time,Idle Time,Point Count,Error Message Count,Error Messages\n");
				init = false;
			}
			
			count = (*data_iter).Error_Message_Count();
			
			text = QString("%1,%2,%3,%4,%5,%6,%7,%8")
						.arg((*data_iter).Name())
						.arg((*data_iter).Revision())
						.arg((*data_iter).Date())
						.arg((*data_iter).Time())
						.arg((*data_iter).Execution_Time())
						.arg((*data_iter).Idle_Time())
						.arg((*data_iter).Executed_Point_Count())
						.arg(count);
			
			for(index = 0;index < count;++index)
			{
				text += QString(",%1").arg((*data_iter).Error_Message(index));
			}
			
			text += '\n';
			
			file.write(text.toLatin1());
		
		}
		
		file.close();
	}
}

void TCMMUsage::Export_Selected_TXT(void)
{
	QString								file_name;
	QFile								file;
	int									index;
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Export Selected to TXT"),
											 d_export_path,
											 QStringLiteral("Text File (*.txt)"));
	
	if(file_name.length())
	{
		file.setFileName(file_name);
		
		if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
		{
			return;
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_export_path = file_name;
			d_export_path.truncate(index);
		}
		
		file.write(d_summary_text->toPlainText().toLatin1());
		
		file.close();
	}
}

void TCMMUsage::Monitor_Started(void)
{
	this->Set_Status_Text(QStringLiteral("INF:  Monitor Started"));
	
	d_monitor_timer.start();
	d_monitor_active = true;
}

void TCMMUsage::Monitor_Finished(void)
{
	int									seconds;

	this->Set_Status_Text(QStringLiteral("INF:  Monitor Stopped"));
	d_monitor_active = false;

	seconds = d_monitor_timer.elapsed() / 1000;

	if(seconds > 0)
	{
		d_database->Increment_Software_Monitor_Time(seconds);
	}
	
	d_program_open = false;
	d_program_executing = false;
	d_active_program_widget->Enable_Collection(false);
}

void TCMMUsage::Update_Timeout(void)
{
	if(d_program_open)
	{
		d_active_program_widget->Increment_Open_Time_MS(d_program_open_timer.restart());
		
		if(d_program_executing)
		{
			d_active_program_widget->Increment_Execution_Time_MS(d_program_execution_timer.restart());
		}
	}
}

void TCMMUsage::Help_About(void)
{
	d_msg_box->setText(QStringLiteral("CMM Usage"));
	d_msg_box->setInformativeText(QStringLiteral("CMM Usage - Version 1.0"));
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	d_msg_box->setIcon(QMessageBox::Information);
	
	d_msg_box->exec();
}

void TCMMUsage::Date_Selection_Changed(
	const QStringList 					&list)
{	
	d_selected_dates = list;
	d_program_data_widget->Set_Data(d_database->Programs(d_selected_dates));
	
	d_selected_programs.clear();
	this->Update_Selected_Text_Data();
	
	d_export_selected_csv_action->setEnabled(false);
	d_export_selected_txt_action->setEnabled(false);
}

void TCMMUsage::Program_Selection_Changed(
	const QStringList 					&list)
{
	d_selected_programs = list;
	
	this->Update_Selected_Text_Data();
	
	d_export_selected_csv_action->setEnabled(d_selected_programs.size() > 0);
	d_export_selected_txt_action->setEnabled(d_selected_programs.size() > 0);
}

void TCMMUsage::Summary_Text_Ready(void)
{
	if(d_summary_text_worker->Has_Valid_Summary())
	{
		d_summary_text->setPlainText(d_summary_text_worker->Text());
		d_view_tab->setCurrentWidget(d_summary_text);
	}
	else
	{
		d_summary_text->setPlainText(QStringLiteral("No summary available."));
	}
	
	d_summary_text_worker->deleteLater();
	d_summary_text_worker = 0;
	
	d_summary_thread->exit(0);	// triggers QThread::finished followed by deleteLater();
	d_summary_thread = 0;
}

#ifndef Q_OS_WIN
void TCMMUsage::Simulation_Timeout(void)
{
	static int							index(0);
	
	if(!(index < d_simulation_events.size()))
	{
		this->Toggle_Monitor();
		index = 0;
		return;
	}
	
	this->Application_Event(d_simulation_events[index],0,0);
	
	++index;
}
#endif

void TCMMUsage::closeEvent(
	QCloseEvent 						*event)
{
	if(d_connection_status != TCMMUsage::NOT_CONNECTED)
	{
		this->Toggle_Monitor();
		d_monitor_process->wait();
	}
	
	d_settings->setValue(QStringLiteral("Mainwindow_Position"), this->pos());
	d_settings->setValue(QStringLiteral("Mainwindow_Size"), this->size());

	d_settings->setValue(QStringLiteral("Export_Path"), d_export_path);

	this->Disconnect_PCDMIS();
	this->Close_Database();
	
	event->accept();
}

bool TCMMUsage::Connect_PCDMIS(void)
{
	TAutomation							*pcdmis_active_program;
	
#ifndef Q_OS_WIN
	return true;
#endif	
			
	// This must be called before anything else	
	TAutomation::InitializeComLibrary();

	if(d_pcdmis.CreateObject(QStringLiteral("PCDLRN.Application")))
	{
		d_pcdmis_path = QDir::fromNativeSeparators(d_pcdmis.GetProperty("Path").AsString());
		
		if(d_pcdmis_path.endsWith('/'))
		{
			d_pcdmis_path.chop(1);
		}
						
		d_pcdmis_application_events = d_pcdmis.GetSubObject(QStringLiteral("ApplicationEvents"));
		
		if(d_pcdmis_application_events)
		{
			d_pcdmis_application_events->Init_Events(d_pcdmis_path + QStringLiteral("/pcdlrn.tlb"),QStringLiteral("IApplicationEvents"));
			d_pcdmis_application_events->Init_Event_Signals(d_pcdmis_path + QStringLiteral("/appevents.tlb"));
			
			if(d_pcdmis_application_events->Enable_Events())
			{
				connect(d_pcdmis_application_events,&TAutomation::Sink_Signal,this,&TCMMUsage::Application_Event);
				
				d_pcdmis_application_events->Add_Event_Signal(QStringLiteral("OnStartExecution"));
				d_pcdmis_application_events->Add_Event_Signal(QStringLiteral("OnEndExecution"));
				d_pcdmis_application_events->Add_Event_Signal(QStringLiteral("OnOpenPartProgram"));
				d_pcdmis_application_events->Add_Event_Signal(QStringLiteral("OnClosePartProgram"));
//				d_pcdmis_application_events->Add_Event_Signal(QStringLiteral("OnUpdateStatusMessage"));
			}
			else
			{
				this->Set_Status_Text(QStringLiteral("ERR:  PC-DMIS events cannot be enabled."));
				
				TAutomation::Release(d_pcdmis_application_events);
				d_pcdmis_application_events = 0;
				
				d_pcdmis.ReleaseObject();
				return false;
			}
			
			pcdmis_active_program = d_pcdmis.GetSubObject(QStringLiteral("ActivePartProgram"));
			
			if(pcdmis_active_program && !pcdmis_active_program->IsNullObject())
			{
				this->Open_PCDMIS_Program(pcdmis_active_program);
				pcdmis_active_program->ReleaseObject();
				pcdmis_active_program = 0;
			}
		}
		else
		{
			this->Set_Status_Text(QStringLiteral("ERR:  PC-DMIS IApplicationEvent information not available."));
			
			d_pcdmis.ReleaseObject();
			return false;
		}
	}
	else
	{
		this->Set_Status_Text(d_pcdmis.GetLastErrorText());
		return false;
	}

	return true;
}

void TCMMUsage::Disconnect_PCDMIS(void)
{
	this->Disconnect_PCDMIS_Machine_Events();
	
	if(d_pcdmis_application_events)
	{
		disconnect(d_pcdmis_application_events,0,0,0);
		d_pcdmis_application_events->ReleaseObject();
		d_pcdmis_application_events->deleteLater();
		d_pcdmis_application_events = 0;
	}

	if(d_pcdmis.DispatchInterface())
	{
		d_pcdmis.ReleaseObject();
		TAutomation::UninitializeComLibrary();
	}
}
bool TCMMUsage::Open_PCDMIS_Program(
	TAutomation 						* const program)
{
	QString								name;
	QString								revision;
	QString								text;
	bool								machine_online(false);

	if(program && !program->IsNullObject())
	{
		name = program->GetProperty("PartName").AsString();
		revision = program->GetProperty("RevisionNumber").AsString();
		
		TAutomation machine(program->GetProperty("ActiveMachine").AsObjectInterface());
		
		if(!machine.IsNullObject())
		{
			text = machine.GetProperty("Name").AsString();
			machine.ReleaseObject();
			
			if(text.startsWith(QStringLiteral("CMM"),Qt::CaseInsensitive))
			{
				machine_online = true;
				this->Connect_PCDMIS_Machine_Events(program);
			}
		}
	}
	
#ifndef Q_OS_WIN
	name = QStringLiteral("Program Name");
	revision = QStringLiteral("Rev 01");
	machine_online = true;
#endif

	if(machine_online)
	{
		d_active_program_widget->Reset();
		d_active_program_widget->Add_Message(QStringLiteral("INF:  Open part program - Online"));
		
		d_program_open = true;
		d_program_open_timer.start();
		
		if(name.length())
		{
			d_active_program_widget->Add_Message(QStringLiteral("INF:  Creating new database entry"));

			d_active_database_item.Set_Name(name);
			d_active_database_item.Set_Revision(revision);
			d_active_database_item.Set_Date_Time(QDateTime::currentDateTime());
						
			d_active_program_widget->Enable_Collection(true);
			d_active_program_widget->Set_Name(name);
			d_active_program_widget->Set_Revision(revision);
			d_active_program_widget->Set_Date(d_active_database_item.Date());
			
			d_view_tab->setCurrentWidget(d_active_program_widget);
			
			return true;
		}
	}
	else
	{
		d_active_program_widget->Add_Message(QStringLiteral("WAR:  Ignoring offline part program."));
	}

	return false;
}

bool TCMMUsage::Close_PCDMIS_Program(
	TAutomation 						* const program)
{
	QString								text;
	bool								machine_online(false);

	if(program && !program->IsNullObject())
	{
		TAutomation machine(program->GetProperty("ActiveMachine").AsObjectInterface());
		
		if(!machine.IsNullObject())
		{
			text = machine.GetProperty("Name").AsString();
			machine.ReleaseObject();
			
			if(text.startsWith(QStringLiteral("CMM"),Qt::CaseInsensitive))
			{
				machine_online = true;
			}
		}
	}
	
#ifndef Q_OS_WIN
	machine_online = true;
#endif
	
	if(machine_online)
	{
		d_active_program_widget->Add_Message(QStringLiteral("INF:  Close part program - Online"));

		this->Disconnect_PCDMIS_Machine_Events();
		
		d_program_open = false;
		d_program_executing = false;
		this->Update_Timeout();

		this->Write_Active_DatabaseItem();
		
		d_active_program_widget->Enable_Collection(false);

		return true;
	}
	else
	{
		d_active_program_widget->Add_Message(QStringLiteral("WAR:  Ignoring offline part program."));
	}

	return false;
}

bool TCMMUsage::Connect_PCDMIS_Machine_Events(
	TAutomation 						* const program)
{
	QString								text;
	
#ifndef Q_OS_WIN
	return true;
#endif	
	
	this->Disconnect_PCDMIS_Machine_Events();
	
	if(program && !program->IsNullObject())
	{
		d_pcdmis_machine_events = program->GetSubObject("ActiveMachine");
		
		if(d_pcdmis_machine_events)
		{
			if(!d_pcdmis_machine_events->Init_Events(d_pcdmis_path + QStringLiteral("/pcdlrn.tlb"),QStringLiteral("IMachineEvents")))
			{
				this->Set_Status_Text(QStringLiteral("ERR:  Init_Events failed"));
			}
			
			if(!d_pcdmis_machine_events->Init_Event_Signals(d_pcdmis_path + QStringLiteral("/machevents.tlb")))
			{
				this->Set_Status_Text(QStringLiteral("ERR:  Init_Event_Signals failed"));
			}
			
			if(d_pcdmis_machine_events->Enable_Events())
			{
				connect(d_pcdmis_machine_events,&TAutomation::Sink_Signal,this,&TCMMUsage::Machine_Event);
				
				d_pcdmis_machine_events->Add_Event_Signal(QStringLiteral("LearnHit"));
				d_pcdmis_machine_events->Add_Event_Signal(QStringLiteral("ExecuteHit"));
				d_pcdmis_machine_events->Add_Event_Signal(QStringLiteral("ErrorMsg"));
			}
			else
			{
				this->Set_Status_Text(QStringLiteral("ERR:  Cannot connect to PC-DMIS machine events..."));
				
				TAutomation::Release(d_pcdmis_machine_events);
				d_pcdmis_machine_events = 0;
				
				return false;
			}
		}
		else
		{
			this->Set_Status_Text(QStringLiteral("ERR:  Cannot connect to PC-DMIS active machine..."));
			return false;
		}
		
		return true;
	}

	return false;
}

void TCMMUsage::Disconnect_PCDMIS_Machine_Events(void)
{
	if(d_pcdmis_machine_events)
	{
		disconnect(d_pcdmis_machine_events,0,0,0);
		d_pcdmis_machine_events->ReleaseObject();
		d_pcdmis_machine_events->deleteLater();
		d_pcdmis_machine_events = 0;
	}
}

bool TCMMUsage::Open_Database(
	const QString 						&file_name)
{
	if(d_database->Is_Connected())
	{
		d_database->Disconnect();
	}
	
	if(d_database->Connect(file_name))
	{
		if(d_database->Initialize())
		{
			return true;
		}
		else
		{
			this->Set_Status_Text(QStringLiteral("ERR:  Cannot initialize database."));
		}
	}
	else
	{
		this->Set_Status_Text(QStringLiteral("ERR:  Cannot connect to database."));
	}
	
	return false;
}

void TCMMUsage::Close_Database(void)
{
	d_database->Disconnect();
}

void TCMMUsage::Write_Active_DatabaseItem(void)
{
	int									execution_time;
	int									idle_time;
	
	if(d_active_database_item.Date_Time_Valid())
	{
		idle_time = d_active_program_widget->Idle_Time();
		execution_time = d_active_program_widget->Execution_Time();
		
		d_active_database_item.Set_Execution_Time(execution_time);
		d_active_database_item.Set_Idle_Time(idle_time);

		d_database->Add_Database_Item(d_active_database_item);
		
		d_date_data_widget->Set_Data(d_database->Dates());
		d_program_data_widget->Set_Data(d_database->Programs(QStringList()));
		
		d_active_program_widget->Add_Message(QStringLiteral("INF:  Database entry written."));
		
		d_active_database_item.Set_Date_Time(QDateTime());	// invalid date time
	}
}

void TCMMUsage::Update_Selected_Text_Data(void)
{
	QStringList							dates;
	
	if(d_summary_text_worker)
	{
		return;
	}
	
	d_summary_text->setPlainText(QStringLiteral("Updating summary..."));
	
	d_summary_thread = new QThread();
	
	if(d_selected_dates.size())
	{
		dates = d_selected_dates;
	}
	else
	{
		dates = d_database->Dates();
	}
	
	d_summary_text_worker = new TSummaryText(d_database,dates,d_selected_programs);
	d_summary_text_worker->moveToThread(d_summary_thread);

	connect(d_summary_thread,&QThread::started,d_summary_text_worker,&TSummaryText::Generate_Summary_Text);
	connect(d_summary_thread,&QThread::finished,d_summary_thread,&QThread::deleteLater);
	connect(d_summary_text_worker,&TSummaryText::Finished,this,&TCMMUsage::Summary_Text_Ready);

	d_summary_thread->start();
}

void TCMMUsage::Set_Status_Text(
	const QString						&text)
{
	d_status_bar->clearMessage();
	d_status_bar->showMessage(text);
}



