/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QVariant>
#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QtSql/QSqlField>
#include <QtSql/QSqlDriver>
#include <QDate>
#include <set>
#include <assert.h>

#include <QFile>
#include <QDataStream>

#include "database.h"

TDatabase::TDatabase(void)
{
	d_database_instance_name = QStringLiteral("cmmusage_database");
	d_database = 0;
}

TDatabase::~TDatabase(void)
{
	this->Disconnect();
}

bool TDatabase::Is_Connected(void) const
{
	if(!d_database)
	{
		return false;
	}
	
	return d_database->isOpen();
}

bool TDatabase::Program_Exists(
	const QString						&name,
	const QString						&date,
	const QString						&time) const
{
	int									id;
	
	id = this->Program_Id(name,date,time);
	
	if(id < 0)
	{
		return false;
	}
	
	return true;
}

bool TDatabase::Program_Exists(
	const QString 						&name,
	const QDateTime 					&date_time) const
{
	QString								date;
	QString								time;
	
	if(!date_time.isValid())
	{
		return false;
	}
	
	date = this->Date_To_String(date_time.date());
	time = this->Time_To_String(date_time.time());
	
	return this->Program_Exists(name,date,time);
}

QStringList TDatabase::Dates(void) const
{
	QStringList							list;
		
	QSqlQuery							query(*d_database);

	if(query.exec(QStringLiteral("SELECT date FROM program")))
	{
		while(query.next())
		{
			list.push_back(query.value(0).toString());
		}
		
		list.removeDuplicates();
	}
	
	return list;
}

QStringList TDatabase::Programs(
	const QStringList 					&dates) const
{
	QStringList							list;
	QStringList::const_iterator			iter;
	
	QSqlQuery							query(*d_database);

	if(dates.size())
	{
		for(iter = dates.begin();iter != dates.end();++iter)
		{
			if(query.exec(QString("SELECT name FROM program WHERE date='%1'").arg(*iter)))
			{
				while(query.next())
				{
					list.push_back(query.value(0).toString());
				}
			}
		}
	}
	else
	{
		if(query.exec(QStringLiteral("SELECT name FROM program")))
		{
			while(query.next())
			{
				list.push_back(query.value(0).toString());
			}
		}
	}
	
	list.removeDuplicates();

	return list;
}

QStringList TDatabase::Program_Times(
	const QString 						&name,
	const QString 						&date) const
{
	QStringList							list;
	
	QSqlQuery							query(*d_database);

	if(query.exec(QString("SELECT time FROM program WHERE name='%1' AND date='%2'").arg(name).arg(date)))
	{
		while(query.next())
		{
			list.push_back(query.value(0).toString());
		}
	}

	return list;
}

std::vector<TDatabaseItem> TDatabase::Program_Data(
	const QString 						&name,
	const QString 						&date)
{
	std::vector<TDatabaseItem>			items;
	TDatabaseItem						item;
	QString								time;
	std::vector<int>					program_ids;
	std::vector<int>::iterator			id_iter;
	int									cntr;
		
	QSqlQuery							query(*d_database);

	item.Set_Name(name);
	
	if(query.exec(QString("SELECT program_id,time,revision FROM program WHERE name='%1' AND date='%2'").arg(name).arg(date)))
	{
		while(query.next())
		{
			program_ids.push_back(query.value(0).toInt());
			
			time = query.value(1).toString();
			item.Set_Revision(query.value(2).toString());
			
			item.Set_Date_Time(this->String_To_DateTime(date,time));
			
			items.push_back(item);
		}
	}
	
	assert(program_ids.size() == items.size());

	for(id_iter = program_ids.begin(),cntr = 0;id_iter != program_ids.end();++id_iter,++cntr)
	{
		if(query.exec(QString("SELECT execution_time,idle_time,point_count FROM program_data WHERE program_id=%1").arg(*id_iter)))
		{
			if(query.next())
			{
				items[cntr].Set_Execution_Time(query.value(0).toInt());
				items[cntr].Set_Idle_Time(query.value(1).toInt());
				items[cntr].Set_Executed_Point_Count(query.value(2).toInt());
			}
		}
	}
	
	for(id_iter = program_ids.begin(),cntr = 0;id_iter != program_ids.end();++id_iter,++cntr)
	{
		if(query.exec(QString("SELECT message FROM program_message WHERE program_id=%1").arg(*id_iter)))
		{
			while(query.next())
			{
				items[cntr].Add_Error_Message(query.value(0).toString());
			}
		}
	}
	
	return items;
}

TDatabaseItem TDatabase::Program_Data(
	const QString 						&name,
	const QString 						&date,
	const QString 						&time,
	bool								* const valid)
{
	TDatabaseItem						item;
	int									program_id(-1);
	
	assert(valid);
	*valid = false;
	
	QSqlQuery							query(*d_database);

	item.Set_Name(name);
	item.Set_Date_Time(this->String_To_DateTime(date,time));
	
	if(query.exec(QString("SELECT program_id,revision FROM program WHERE name='%1' AND date='%2' AND time='%3'").arg(name).arg(date).arg(time)))
	{
		if(query.next())
		{
			program_id = query.value(0).toInt();
			item.Set_Revision(query.value(1).toString());
		}
	}

	if(program_id < 0)
	{
		return item;
	}
	
	if(query.exec(QString("SELECT execution_time,idle_time,point_count FROM program_data WHERE program_id=%1").arg(program_id)))
	{
		if(query.next())
		{
			item.Set_Execution_Time(query.value(0).toInt());
			item.Set_Idle_Time(query.value(1).toInt());
			item.Set_Executed_Point_Count(query.value(2).toInt());
			
			*valid = true;
		}
	}
	
	if(*valid)
	{
		if(query.exec(QString("SELECT message FROM program_message WHERE program_id=%1").arg(program_id)))
		{
			while(query.next())
			{
				item.Add_Error_Message(query.value(0).toString());
			}
		}
	}
	
	return item;
}

int TDatabase::Software_Monitor_Time(
	bool 								* const valid)
{
	int									value(0);
	
	assert(valid);
	*valid = false;
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QStringLiteral("SELECT software_monitor_time FROM usage")))
	{
		if(query.next())
		{
			value = query.value(0).toInt();
			*valid = true;
		}
	}

	return value;
}

int TDatabase::Software_Run_Time(
	bool 								* const valid)
{
	int									value(0);

	assert(valid);
	*valid = false;
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QStringLiteral("SELECT software_run_time FROM usage")))
	{
		if(query.next())
		{
			value = query.value(0).toInt();
			*valid = true;
		}
	}

	return value;
}

bool TDatabase::Connect(
	const QString						&file_name)
{
	QStringList							list;
	QString								text;
	QList<QString>::const_iterator		iter;
	
	if(!QSqlDatabase::isDriverAvailable(QStringLiteral("QSQLITE")))
	{
		d_last_error = QStringLiteral("Database driver 'QSQLITE' not available");
		return false;
	}
	
	if(QSqlDatabase::contains(d_database_instance_name))
	{
		d_database = new QSqlDatabase(QSqlDatabase::database(d_database_instance_name));
	}
	else
	{
		d_database = new QSqlDatabase(QSqlDatabase::addDatabase(QStringLiteral("QSQLITE"),d_database_instance_name));
	}
	
	d_database->setDatabaseName(file_name);
	
	if(!d_database->open())
	{
		text = QString("Cannot open database.\n\tDatabase Error: '%1'\n\tDriver Error: '%2'\n")
									.arg(d_database->lastError().databaseText())
									.arg(d_database->lastError().driverText());
		
		text += QString("\t\tDatabase File Name: '%1'\n").arg(d_database_file_name);
		
		list = QSqlDatabase::drivers();
		
		text += QString("\tAvailable database drivers:\n");
		for(iter = list.begin();iter != list.end();++iter)
		{
			text += QString("\t\t%1\n").arg(*iter);
		}
		
		QSqlDatabase::removeDatabase(d_database_instance_name);
		
		delete d_database;
		d_database = 0;

		d_last_error = text;
		return false;
	}
	
	d_database_file_name = file_name;

	return true;
}

bool TDatabase::Initialize(void)
{
	QStringList							list;
	QString								text;
	QList<QString>::const_iterator		iter;

	if(!d_database)
	{
		d_last_error = QStringLiteral("Database not open");
		return false;
	}
	
	if(!d_database->isOpen())
	{
		d_last_error = QStringLiteral("Database not open");
		return false;
	}
	
	// create tables if necessary
	QSqlQuery							query(*d_database);
	
	list = d_database->tables();

	if(!list.contains("program",Qt::CaseInsensitive))
	{
		text = QStringLiteral("CREATE TABLE program(program_id INTEGER PRIMARY KEY AUTOINCREMENT,name TEXT,revision TEXT,date TEXT,time TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'program(...)'");
			return false;
		}
	}

	if(!list.contains("program_data",Qt::CaseInsensitive))
	{
		text = QStringLiteral("CREATE TABLE program_data(program_id INTEGER,execution_time INTEGER,idle_time INTEGER,point_count INTEGER)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'program_data(...)'");
			return false;
		}
	}
	
	if(!list.contains("program_message",Qt::CaseInsensitive))
	{
		text = QStringLiteral("CREATE TABLE program_message(program_id INTEGER,message TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'program_message(...)'");
			return false;
		}
	}
	
	if(!list.contains("usage",Qt::CaseInsensitive))
	{
		text = QStringLiteral("CREATE TABLE usage(software_monitor_time INTEGER,software_run_time INTEGER)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'usage(...)'");
			return false;
		}
				
		text = QStringLiteral("INSERT INTO usage(software_monitor_time,software_run_time) VALUES(0,0)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot initialize table 'usage(...)'");
			return false;
		}
	}

	return true;
}

void TDatabase::Disconnect(void)
{
	if(d_database)
	{
		d_database->close();
		
		delete d_database;
		d_database = 0;

		if(QSqlDatabase::contains(d_database_instance_name))
		{
			QSqlDatabase::removeDatabase(d_database_instance_name);
		}
	}
}

void TDatabase::Clear(void)
{
	if(d_database)
	{
		QSqlQuery							query(*d_database);
		
		query.exec(QStringLiteral("DELETE FROM program"));
		query.exec(QStringLiteral("DELETE FROM program_data"));
		query.exec(QStringLiteral("DELETE FROM program_message"));
	}
}

bool TDatabase::Add_Database_Item(
	const TDatabaseItem					&database_item)
{
	QString								command;
	int									program_id;
	int									count;
	int									cntr;
	QDateTime							date_time;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}

	QSqlQuery							query(*d_database);
	
	date_time = database_item.Date_Time();
	
	if(!database_item.Date_Time_Valid())
	{
		d_last_error = QStringLiteral("Date_Time invalid");
		return false;
	}

	if(!this->Program_Exists(database_item.Name(),database_item.Date_Time()))
	{
		command = QString("INSERT INTO program (name,revision,date,time) VALUES ('%1','%2','%3','%4')")
							.arg(database_item.Name())
							.arg(database_item.Revision())
							.arg(this->Date_To_String(date_time.date()))
							.arg(this->Time_To_String(date_time.time()));

		if(!query.exec(command))
		{
			d_last_error = QString("Cannot add data to table 'program'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}
	else
	{
		d_last_error = QStringLiteral("Cannot add data to table 'program'.  Reason: Program entry already exists");
		return false;
	}
	
	program_id = this->Program_Id(database_item.Name(),database_item.Date_Time());

	command = QString("INSERT INTO program_data (program_id,execution_time,idle_time,point_count) VALUES (%1,%2,%3,%4)")
						.arg(program_id)
						.arg(database_item.Execution_Time())
						.arg(database_item.Idle_Time())
						.arg(database_item.Executed_Point_Count());

	if(!query.exec(command))
	{
		d_last_error = QString("Cannot add data to table 'program_data'.  Reason: %1").arg(query.lastError().text());
		return false;
	}
	
	count = database_item.Error_Message_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		command = QString("INSERT INTO program_message (program_id,message) VALUES (%1,'%2')")
							.arg(program_id)
							.arg(database_item.Error_Message(cntr));
		
		if(!query.exec(command))
		{
			d_last_error = QString("Cannot add data to table 'program_message'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}

	return true;
}

bool TDatabase::Increment_Software_Monitor_Time(
	const int 							seconds)
{
	QString								command;
	int									value;
	bool								valid;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}

	value = this->Software_Monitor_Time(&valid);
	value += seconds;
	
	QSqlQuery							query(*d_database);
	
	command = QString("UPDATE usage SET software_monitor_time = %1").arg(value);

	if(!query.exec(command))
	{
		d_last_error = QString("Cannot add data to table 'usage'.  Reason: %1").arg(query.lastError().text());
		return false;
	}

	return true;
}

bool TDatabase::Increment_Software_Run_Time(
	const int 							seconds)
{
	QString								command;
	int									value;
	bool								valid;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}

	value = this->Software_Run_Time(&valid);
	value += seconds;
	
	QSqlQuery							query(*d_database);
	
	command = QString("UPDATE usage SET software_run_time = %1").arg(value);

	if(!query.exec(command))
	{
		d_last_error = QString("Cannot add data to table 'usage'.  Reason: %1").arg(query.lastError().text());
		return false;
	}

	return true;
}

int TDatabase::Program_Id(
	const QString 						&name,
	const QString						&date,
	const QString						&time) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	if(!name.length())
	{
		d_last_error = QStringLiteral("Invalid name");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(date.length() && time.length())
	{
		if(query.exec(QString("SELECT program_id FROM program WHERE name='%1' AND date='%2' AND time='%3'")
					  .arg(name)
					  .arg(date)
					  .arg(time)))
		{
			if(query.next())
			{
				return query.value(0).toInt();
			}
		}
	}
	else if(date.length())
	{
		if(query.exec(QString("SELECT program_id FROM program WHERE name='%1' AND date='%2'")
					  .arg(name)
					  .arg(date)))
		{
			if(query.next())
			{
				return query.value(0).toInt();
			}
		}
	}
	else if(time.length())
	{
		if(query.exec(QString("SELECT program_id FROM program WHERE name='%1' AND time='%2'")
					  .arg(name)
					  .arg(time)))
		{
			if(query.next())
			{
				return query.value(0).toInt();
			}
		}
	}
	else
	{
		if(query.exec(QString("SELECT program_id FROM program WHERE name='%1'")
					  .arg(name)))
		{
			if(query.next())
			{
				return query.value(0).toInt();
			}
		}
	}


	d_last_error = QStringLiteral("Program not found.");
	return -1;
}

int TDatabase::Program_Id(
	const QString 						&name,
	const QDateTime 					&date_time) const
{
	QString								date;
	QString								time;

	if(!date_time.isValid())
	{
		return -1;
	}
	
	date = this->Date_To_String(date_time.date());
	time = this->Time_To_String(date_time.time());
	
	return this->Program_Id(name,date,time);
}

QString TDatabase::Date_To_String(
	const QDate 						&date)
{
	QString								date_string;
	
	date_string = date.toString(QStringLiteral("yyyy MMM dd"));
	
	return date_string;
}

QString TDatabase::Time_To_String(
	const QTime 						&time)
{
	QString								time_string;
	
	time_string = time.toString(QStringLiteral("hh mm ss ap"));
	
	return time_string;
}

QDate TDatabase::String_To_Date(
	const QString 						&date)
{
	return QDate::fromString(date,QStringLiteral("yyyy MMM dd"));
}

QTime TDatabase::String_To_Time(
	const QString 						&time)
{
	return QTime::fromString(time,QStringLiteral("hh mm ss ap"));
}

QDateTime TDatabase::String_To_DateTime(
	const QString 						&date,
	const QString 						&time)
{
	QDateTime							date_time;
	
	date_time.setDate(TDatabase::String_To_Date(date));
	date_time.setTime(TDatabase::String_To_Time(time));
	
	return date_time;
}



