/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QApplication>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QListView>
#include <QVBoxLayout>
#include <QStringListModel>
#include <QSortFilterProxyModel>
#include <QItemSelectionModel>
#include <QModelIndex>
#include <QModelIndexList>
#include <QPixmap>
#include <QFont>
#include <assert.h>

#include "datawidget.h"

TDataWidget::TDataWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QHBoxLayout							*search_hlayout;
	QLabel								*search_icon_label;
	QVBoxLayout							*widget_vlayout;
	QFont								title_font;
	
	title_font.setPointSize(14);
	title_font.setUnderline(true);

	widget_vlayout = new QVBoxLayout(this);
	widget_vlayout->setContentsMargins(0,0,0,0);
	
	d_title_label = new QLabel(this);
	d_title_label->setFont(title_font);
	widget_vlayout->addWidget(d_title_label);

	search_hlayout = new QHBoxLayout();

	search_icon_label = new QLabel(this);
	search_hlayout->addWidget(search_icon_label);

	d_search_edit = new QLineEdit(this);
	search_hlayout->addWidget(d_search_edit);
	widget_vlayout->addLayout(search_hlayout);

	d_data_list = new QListView(this);
	d_data_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	widget_vlayout->addWidget(d_data_list);
	
	d_string_list_model = new QStringListModel(this);
	
	d_sort_filter_proxy_model = new QSortFilterProxyModel(this);
	d_sort_filter_proxy_model->setSourceModel(d_string_list_model);
	d_sort_filter_proxy_model->setFilterCaseSensitivity(Qt::CaseInsensitive);

	d_data_list->setModel(d_sort_filter_proxy_model);
	
	search_icon_label->setPixmap(QPixmap(QStringLiteral(":/icons/search_x16.png")));
	
	d_title_label->setText(QStringLiteral("title"));
	
	connect(d_search_edit,&QLineEdit::textEdited,d_sort_filter_proxy_model,static_cast<void (QSortFilterProxyModel::*)(const QString&)>(&QSortFilterProxyModel::setFilterRegExp));	// older versions of qt don't use RegularExpression
	
	connect(d_data_list->selectionModel(),&QItemSelectionModel::selectionChanged,this,&TDataWidget::Model_Selection_Changed);
}

TDataWidget::~TDataWidget(void)
{
}

void TDataWidget::Set_Title(
	const QString 						&text)
{
	d_title_label->setText(text);
}

void TDataWidget::Set_Data(
	const QStringList 					&list)
{
	d_string_list_model->setStringList(list);
}

void TDataWidget::Model_Selection_Changed(
	const QItemSelection 				&selected,
	const QItemSelection 				&deselected)
{
	Q_UNUSED(selected);
	Q_UNUSED(deselected);
	
	QModelIndexList						index_list;
	QModelIndexList::const_iterator		iter;
	QStringList							list;
	
	index_list = d_data_list->selectionModel()->selectedIndexes();
	
	for(iter = index_list.begin();iter != index_list.end();++iter)
	{
		list.push_back(d_string_list_model->data((*iter),Qt::DisplayRole).toString());
	}
	
	emit Selection_Changed(list);
}


