/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <assert.h>

#include "summarytext.h"

TSummaryText::TSummaryText(
	TDatabase 							*database,
	const QStringList 					&dates,
	const QStringList 					&programs)
{
	d_database = database;
	d_dates = dates;
	d_programs = programs;
	
	d_has_valid_summary = false;
}

TSummaryText::~TSummaryText(void)
{
}

void TSummaryText::Generate_Summary_Text(void)
{
	TDatabaseItem						data_item;
	std::vector<TDatabaseItem>			data_items;
	std::vector<TDatabaseItem>			data_items_date;
	std::vector<TDatabaseItem>::const_iterator data_iter;
	QStringList							program_dates;
	QStringList::const_iterator			date_iter;
	QStringList::const_iterator			program_iter;
	QStringList							all_error_messages;
	QStringList::const_iterator			message_iter;
	double								dval;
	int									sum_execution_time(0);
	int									sum_idle_time(0);
	int									sum_total_time(0);
	int									sum_point_count(0);
	int									execution_time;
	int									idle_time;
	int									point_count;
	int									ival;
	int									count;
	int									cntr;
	int									minutes;
	int									seconds;
	bool								valid;

	this->Clear_Summary();
	
	if(d_programs.size())
	{
		d_has_valid_summary = true;
	}
	else
	{
		d_has_valid_summary = false;
	}
			
	if(d_dates.size())
	{
		program_dates = d_dates;
	}
	else
	{
		program_dates = d_database->Dates();
	}
	
	for(date_iter = d_dates.begin();date_iter != d_dates.end();++date_iter)
	{
		for(program_iter = d_programs.begin();program_iter != d_programs.end();++program_iter)
		{
			data_items_date = d_database->Program_Data((*program_iter),(*date_iter));
			data_items.insert(data_items.end(),data_items_date.begin(),data_items_date.end());
		}
	}
		
	ival = d_database->Software_Monitor_Time(&valid);
	
	if(valid)
	{
		dval = static_cast<double>(ival) / 60.0;
		minutes = static_cast<int>(dval);
		seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
		this->Add_Summary_Text(QString("Total Monitor Time:      %1:%2 min:sec\n").arg(minutes,10).arg(seconds,2,10,QChar('0')));
	}
	else
	{
		this->Add_Summary_Text(QStringLiteral("Total Monitor Time:               <unavailable>\n"));
	}
	
	ival = d_database->Software_Run_Time(&valid);

	if(valid)
	{
		dval = static_cast<double>(ival) / 60.0;
		minutes = static_cast<int>(dval);
		seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
		this->Add_Summary_Text(QString("Total Software Run Time: %1:%2 min:sec\n").arg(minutes,10).arg(seconds,2,10,QChar('0')));
	}
	else
	{
		this->Add_Summary_Text(QStringLiteral("Total Software Run Time:          <unavailable>\n"));
	}
	this->Add_Summary_Text(QStringLiteral("---------------------------------------------------------------------------\n"));


	this->Add_Summary_Text(QStringLiteral("     Selected Program Details:\n"));
	this->Add_Summary_Text(QStringLiteral("---------------------------------------------------------------------------\n"));

	for(data_iter = data_items.begin();data_iter != data_items.end();++data_iter)
	{
		execution_time = (*data_iter).Execution_Time();
		idle_time = (*data_iter).Idle_Time();
		point_count = (*data_iter).Executed_Point_Count();
		
		sum_execution_time += execution_time;
		sum_idle_time += idle_time;
		sum_point_count += point_count;
		sum_total_time += (execution_time + idle_time);

		this->Add_Summary_Text(QString("Program Name:                     %1\n").arg((*data_iter).Name()));
		this->Add_Summary_Text(QString("Program Revision:                 %1\n").arg((*data_iter).Revision()));
		this->Add_Summary_Text(QString("Execution Date:                   %1\n").arg((*data_iter).Date_Time().toString(QStringLiteral("yyyy MMM dd : hh mm ss ap"))));

		dval = static_cast<double>(execution_time + idle_time) / 60.0;
		minutes = static_cast<int>(dval);
		seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
		this->Add_Summary_Text(QString("Total Time:            %1:%2 min:sec\n").arg(minutes,12).arg(seconds,2,10,QChar('0')));
		
		dval = static_cast<double>(execution_time) / 60.0;
		minutes = static_cast<int>(dval);
		seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
		this->Add_Summary_Text(QString("Execution Time:        %1:%2 min:sec\n").arg(minutes,12).arg(seconds,2,10,QChar('0')));
		
		dval = static_cast<double>(idle_time) / 60.0;
		minutes = static_cast<int>(dval);
		seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
		this->Add_Summary_Text(QString("Idle Time:             %1:%2 min:sec\n").arg(minutes,12).arg(seconds,2,10,QChar('0')));
		
		this->Add_Summary_Text(QString("Point Count:           %1\n").arg(point_count,15));
		
		count = (*data_iter).Error_Message_Count();
		
		if(count)
		{
			this->Add_Summary_Text(QStringLiteral("Controller Messages During Execution:\n"));
		}
		
		for(cntr = 0;cntr < count;++cntr)
		{
			all_error_messages.push_back((*data_iter).Error_Message(cntr));
			this->Add_Summary_Text(QString(" %1\n").arg((*data_iter).Error_Message(cntr)));
		}
		this->Add_Summary_Text(QStringLiteral("---------------------------------------------------------------------------\n"));
	}
	
	this->Add_Summary_Text(QStringLiteral("     Selected Program Summary:\n"));
	this->Add_Summary_Text(QStringLiteral("---------------------------------------------------------------------------\n"));
	
	this->Add_Summary_Text(QString("Program Count:         %1\n").arg(data_items.size(),15));

	dval = static_cast<double>(sum_total_time) / 60.0;
	minutes = static_cast<int>(dval);
	seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
	this->Add_Summary_Text(QString("Program Time:    %1:%2 min:sec\n").arg(minutes,18).arg(seconds,2,10,QChar('0')));
	
	dval = static_cast<double>(sum_execution_time) / 60.0;
	minutes = static_cast<int>(dval);
	seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
	this->Add_Summary_Text(QString("Execution Time:  %1:%2 min:sec\n").arg(minutes,18).arg(seconds,2,10,QChar('0')));
	
	dval = static_cast<double>(sum_idle_time) / 60.0;
	minutes = static_cast<int>(dval);
	seconds = static_cast<int>((dval - static_cast<double>(minutes)) * 60.0);
	this->Add_Summary_Text(QString("Idle Time:       %1:%2 min:sec\n").arg(minutes,18).arg(seconds,2,10,QChar('0')));
	this->Add_Summary_Text(QString("Point Count:     %1\n").arg(sum_point_count,21));
	
	all_error_messages.removeDuplicates();
	
	this->Add_Summary_Text(QStringLiteral("\nUnique Error Messages:\n"));
	this->Add_Summary_Text(QStringLiteral("---------------------------------------------------------------------------\n"));

	for(message_iter = all_error_messages.begin();message_iter != all_error_messages.end();++message_iter)
	{
		this->Add_Summary_Text(QString(" %1\n").arg(*message_iter));
	}
	
	emit Finished();
}

void TSummaryText::Clear_Summary(void)
{
	d_summary_text.clear();
}

void TSummaryText::Add_Summary_Text(
	const QString						&text)
{
	d_summary_text.append(text);
}
