/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QString>
#include <QStringList>
#include <QUuid>
#include <QDir>
#include <QThread>

#include <assert.h>

#ifndef Q_OS_WIN

// forward class declaration
class IUnknown;
class IDispatch;

#endif

#include "automation.h"

#ifdef Q_OS_WIN

class TAutomationEvent : public IDispatch
{
public:
	TAutomationEvent(TAutomation *parent,IID iid);
	virtual ~TAutomationEvent(void);
	
	// implementation of IUnknown
	ULONG __stdcall AddRef(void);
	ULONG __stdcall Release(void);
	HRESULT __stdcall QueryInterface(REFIID riid, LPVOID *ppv);
	
	// implementation of IDispatch
    HRESULT __stdcall GetIDsOfNames(REFIID, wchar_t **, unsigned int, unsigned long, long *);
    HRESULT __stdcall GetTypeInfoCount(unsigned int *);
    HRESULT __stdcall GetTypeInfo(UINT, LCID, ITypeInfo **);
    HRESULT __stdcall Invoke(DISPID dispIdMember,
                            REFIID riid,
                            LCID,
                            WORD wFlags,
                            DISPPARAMS *pDispparam_s,
                            VARIANT*,
                            EXCEPINFO*,
                            UINT*);
	
private:
	LONG								d_ref;	
	IID									d_sink_iid;
	TAutomation							*d_parent;
};

TAutomationEvent::TAutomationEvent(
	TAutomation 						*parent,
	IID 								iid)
:	d_ref(0),
	d_sink_iid(iid),
	d_parent(parent)
{
}

TAutomationEvent::~TAutomationEvent(void)
{
}
ULONG TAutomationEvent::AddRef(void) 
{
	return ::InterlockedIncrement(&d_ref);
}
	
ULONG TAutomationEvent::Release(void) 
{      
	int count;  
	
	count = ::InterlockedDecrement(&d_ref);
		
    if (count == 0)
    {
        delete this;
	}
		
    return count;
}

HRESULT TAutomationEvent::QueryInterface(
	REFIID 								riid,
	LPVOID 								*ppv) 
{
	if(!ppv)
	{
		return E_INVALIDARG;
	}
	
	if(riid == IID_IUnknown || riid == IID_IDispatch)
	{
		*ppv = (LPVOID)this;
		this->AddRef();
		
		return NOERROR;
	}
	
	if(riid == d_sink_iid)
	{
		*ppv = (LPVOID)this;
		this->AddRef();
		
		return NOERROR;
	}
		
	return E_NOINTERFACE;
}

HRESULT TAutomationEvent::GetIDsOfNames(
	REFIID 								riid,
	wchar_t 							**rgszNames,
	unsigned int						cNames,
	unsigned long						lcid,
	long 								*rgDispId)
{
	Q_UNUSED(riid);
	Q_UNUSED(rgszNames);
	Q_UNUSED(cNames);
	Q_UNUSED(lcid);
	Q_UNUSED(rgDispId);
	
	return E_NOTIMPL; 
}

HRESULT TAutomationEvent::GetTypeInfoCount(
	unsigned int 						*pctinfo)
{
	Q_UNUSED(pctinfo);

	return E_NOTIMPL; 
}

HRESULT TAutomationEvent::GetTypeInfo(
	UINT								iTInfo,
	LCID								lcid, 
	ITypeInfo 							**ppTInfo)
{
	Q_UNUSED(iTInfo);
	Q_UNUSED(lcid);
	Q_UNUSED(ppTInfo);
	
	return E_NOTIMPL;
}
  
HRESULT TAutomationEvent::Invoke(
	DISPID 								dispatch_member_id,
	REFIID 								riid,
	LCID								lcid,
	WORD 								wFlags,
	DISPPARAMS 							*dispatch_parameters,
	VARIANT								*return_variant,
	EXCEPINFO							*pExcepInfo,
	UINT								*puArgErr)    
{
	Q_UNUSED(lcid);
	Q_UNUSED(return_variant);
	Q_UNUSED(pExcepInfo);
	Q_UNUSED(puArgErr);
	
	assert(d_parent);
	    	
	// verify input
	if (riid != IID_NULL)
	{
		return DISP_E_UNKNOWNINTERFACE;
	}
	
	if (!(wFlags & DISPATCH_METHOD))
	{
		return DISP_E_MEMBERNOTFOUND;
	}
		
	d_parent->Send_Signal(dispatch_member_id, 
						static_cast<int>(dispatch_parameters->cArgs), 
						dispatch_parameters->rgvarg);
	
    return S_OK;
}
#endif

TAutomation::TArg::TArg(void)
{
	d_value_type = TArg::ARG_EMPTY;
}

TAutomation::TArg::TArg(
    const TAutomation::TArg 			&rhs)
{
	if(&rhs != this)
	{
		d_value_type = rhs.d_value_type;
		d_value = rhs.d_value;
		d_text = rhs.d_text;
		d_text_array = rhs.d_text_array;
	}
}

TAutomation::TArg& TAutomation::TArg::operator=(
    const TAutomation::TArg 			&rhs)
{
	if(&rhs != this)
	{
		d_value_type = rhs.d_value_type;
		d_value = rhs.d_value;
		d_text = rhs.d_text;
		d_text_array = rhs.d_text_array;
	}

	return *this;
}

TAutomation::TArg::TArg(
    const bool 							val)
{
	d_value_type = ARG_BOOL;
	d_value.d_boolean = val;
}

TAutomation::TArg::TArg(
    const char 							*text)
{
	d_value_type = ARG_STRING;
	d_text = QString::fromLatin1(text);
}

TAutomation::TArg::TArg(
    const QString 						&text)
{
	d_value_type = ARG_STRING;
	d_text = text;
}


TAutomation::TArg::TArg(
    const QStringList					&string_list)
{
	d_value_type = ARG_STRINGLIST;
	d_text_array = string_list;
}

TAutomation::TArg::TArg(
    const float 						&val)
{
	d_value_type = ARG_FLOAT;
	d_value.d_float = val;
}

TAutomation::TArg::TArg(
    const double 						&val)
{
	d_value_type = ARG_DOUBLE;
	d_value.d_double = val;
}

TAutomation::TArg::TArg(
    const int 							val)
{
	d_value_type = ARG_INTEGER;
	d_value.d_integer = val;
}

TAutomation::TArg::TArg(
    double 								*val)
{
	d_value_type = ARG_PTR_DOUBLE;
	d_value.d_pointer_double = val;
}

TAutomation::TArg::TArg(
	int									*val)
{
	d_value_type = ARG_PTR_INTEGER;
	d_value.d_pointer_integer = val;
}

TAutomation::TArg::TArg(
    IDispatch							*val)
{
	d_value_type = ARG_PTR_DISPATCH;
	d_value.d_dispatch_interface = val;
}

TAutomation::TArg::~TArg(void)
{

}

bool TAutomation::TArg::AsBool(void) const
{
	if(d_value_type == ARG_BOOL)
	{
		return d_value.d_boolean;
	}

	if(d_value_type == ARG_INTEGER)
	{
		return (d_value.d_integer != 0);
	}

	if(d_value_type == ARG_DOUBLE)
	{
		return (d_value.d_double != 0.0);
	}

	if(d_value_type == ARG_FLOAT)
	{
		return (d_value.d_float != 0.0);
	}

	if(d_value_type == ARG_STRING)
	{
		return !d_text.isEmpty();
	}

	if(d_value_type == ARG_PTR_DOUBLE)
	{
		return (d_value.d_pointer_double != 0);
	}
	
	if(d_value_type == ARG_PTR_INTEGER)
	{
		return (d_value.d_pointer_integer != 0);
	}

	if(d_value_type == ARG_PTR_DISPATCH)
	{
		return (d_value.d_dispatch_interface != 0);
	}

	return false;
}

QString TAutomation::TArg::AsString(void) const
{
	if(d_value_type == ARG_STRING)
	{
		return d_text;
	}

	if(d_value_type == ARG_BOOL)
	{
		return d_value.d_boolean ? QString("true") : QString("false");
	}

	if(d_value_type == ARG_INTEGER)
	{
		return QString().setNum(d_value.d_integer);
	}

	if(d_value_type == ARG_DOUBLE)
	{
		return QString().setNum(d_value.d_double);
	}

	if(d_value_type == ARG_FLOAT)
	{
		return QString().setNum(d_value.d_float);
	}

	return QString();
}

QString TAutomation::TArg::AsStringFromArray(
	const int 							offset) const
{
	if(d_value_type == ARG_STRINGLIST)
	{
		if(offset < d_text_array.size())
		{
			return d_text_array[offset];
		}
	}

	return QString();
}

int TAutomation::TArg::StringArraySize(void) const
{
	if(d_value_type == ARG_STRINGLIST)
	{
		return d_text_array.size();
	}

	return 0;
}


int TAutomation::TArg::AsInteger(void) const
{
	if(d_value_type == ARG_FLOAT)
	{
		return static_cast<int>(d_value.d_float);
	}

	if(d_value_type == ARG_INTEGER)
	{
		return d_value.d_integer;
	}

	if(d_value_type == ARG_BOOL)
	{
		return d_value.d_boolean == 0 ? 0 : 1;
	}

	if(d_value_type == ARG_DOUBLE)
	{
		return static_cast<int>(d_value.d_double);
	}

	if(d_value_type == ARG_STRING)
	{
		return d_text.toInt();
	}
	
	if(d_value_type == ARG_PTR_INTEGER)
	{
		if(d_value.d_pointer_integer)
		{
			return static_cast<int>(*d_value.d_pointer_integer);
		}
		
		return 0;
	}


	return 0;
}

float TAutomation::TArg::AsFloat(void) const
{
	if(d_value_type == ARG_FLOAT)
	{
		return d_value.d_float;
	}

	if(d_value_type == ARG_DOUBLE)
	{
		return static_cast<float>(d_value.d_double);
	}

	if(d_value_type == ARG_BOOL)
	{
		return d_value.d_boolean == 0 ? 0.0 : 1.0;
	}

	if(d_value_type == ARG_INTEGER)
	{
		return static_cast<float>(d_value.d_integer);
	}

	if(d_value_type == ARG_STRING)
	{
		return d_text.toFloat();
	}

	if(d_value_type == ARG_PTR_DOUBLE)
	{
		if(d_value.d_pointer_double)
		{
			return static_cast<float>(*d_value.d_pointer_double);
		}

		return float(0);
	}

	return float(0);
}

double TAutomation::TArg::AsDouble(void) const
{
	if(d_value_type == ARG_FLOAT)
	{
		return static_cast<double>(d_value.d_float);
	}

	if(d_value_type == ARG_DOUBLE)
	{
		return d_value.d_double;
	}

	if(d_value_type == ARG_BOOL)
	{
		return d_value.d_boolean == 0 ? 0.0 : 1.0;
	}

	if(d_value_type == ARG_INTEGER)
	{
		return static_cast<double>(d_value.d_integer);
	}

	if(d_value_type == ARG_STRING)
	{
		return d_text.toDouble();
	}
	
	if(d_value_type == ARG_PTR_INTEGER)
	{
		if(d_value.d_pointer_integer)
		{
			return static_cast<double>(*d_value.d_pointer_integer);
		}
		
		return 0.0;
	}

	if(d_value_type == ARG_PTR_DOUBLE)
	{
		if(d_value.d_pointer_double)
		{
			return *d_value.d_pointer_double;
		}

		return 0.0;
	}

	return 0.0;
}

double* TAutomation::TArg::AsPtrDouble(void) const
{
	if(d_value_type == ARG_PTR_DOUBLE)
	{
		return d_value.d_pointer_double;
	}

	return 0;
}

int* TAutomation::TArg::AsPtrInteger(void) const
{	
	if(d_value_type == ARG_PTR_INTEGER)
	{
		return d_value.d_pointer_integer;
	}
	
	return 0;
}

TAutomation* TAutomation::TArg::AsObject(void) const
{
	if(d_value_type == ARG_PTR_DISPATCH)
	{
		return new TAutomation(d_value.d_dispatch_interface);
	}

	return 0;
}

IDispatch* TAutomation::TArg::AsObjectInterface(void) const
{
	if(d_value_type == ARG_PTR_DISPATCH)
	{
		return d_value.d_dispatch_interface;
	}

	return 0;
}

TAutomation::TAutomation(const QObject *parent)
:	QObject(const_cast<QObject*>(parent)),
 	d_unknown_interface(0),
	d_dispatch_interface(0)
{
#ifdef Q_OS_WIN

	d_automation_event = 0;
	d_automation_event_connection_point = 0;
	d_automation_event_connection_id = 0;

	memset(&d_automation_event_iid,0,sizeof(IID));
	
#endif
	
}

TAutomation::TAutomation(
    const QString						&name,
	const QObject						*parent)
:	QObject(const_cast<QObject*>(parent)),
 	d_unknown_interface(0),
	d_dispatch_interface(0)
{
#ifdef Q_OS_WIN

	d_automation_event = 0;
	d_automation_event_connection_point = 0;
	d_automation_event_connection_id = 0;

	memset(&d_automation_event_iid,0,sizeof(IID));
	
#endif
	
	this->CreateObject(name);
}

TAutomation::TAutomation(
    IDispatch							*dispatch,
	const QObject						*parent)
:	QObject(const_cast<QObject*>(parent)),
 	d_unknown_interface(0),
	d_dispatch_interface(dispatch)
{
#ifdef Q_OS_WIN
	
	d_automation_event = 0;
	d_automation_event_connection_point = 0;
	d_automation_event_connection_id = 0;
	
	memset(&d_automation_event_iid,0,sizeof(IID));
	
#endif
}

TAutomation::~TAutomation(void)
{	
	this->ReleaseObject();
}

QStringList TAutomation::GetRunningObjectNames(void)
{
	QStringList							name_list;

#ifdef Q_OS_WIN

	IBindCtx							*bind_context_interface;
	IRunningObjectTable					*running_object_interface;
	IEnumMoniker						*moniker_enum_interface;
	IMoniker							*moniker_interface;
	IMalloc								*malloc_interface;
	ULONG								actual_retrieved;
	wchar_t								*display_name(0);
	HRESULT								result_status;
	QString								text;

	result_status = CoGetMalloc(1,&malloc_interface);

	if(FAILED(result_status))
	{
		d_last_error = QStringLiteral("GetRunningObjectNames Error: ") + this->Format_Status_Text(result_status);
		return name_list;
	}

	result_status = CreateBindCtx(0,&bind_context_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();

		d_last_error = QStringLiteral("GetRunningObjectNames Error: ") + this->Format_Status_Text(result_status);
		return name_list;
	}

	result_status = bind_context_interface->GetRunningObjectTable(&running_object_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();
		bind_context_interface->Release();

		d_last_error = QStringLiteral("GetRunningObjectNames Error: ") + this->Format_Status_Text(result_status);
		return name_list;
	}

	result_status = running_object_interface->EnumRunning(&moniker_enum_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();
		running_object_interface->Release();
		bind_context_interface->Release();

		d_last_error = QStringLiteral("GetRunningObjectNames Error: ") + this->Format_Status_Text(result_status);
		return name_list;
	}

	moniker_enum_interface->Reset();

	while(moniker_enum_interface->Next(1,&moniker_interface,&actual_retrieved) == S_OK)
	{
		moniker_interface->GetDisplayName(bind_context_interface,0,&display_name);

		text = QString::fromWCharArray(static_cast<wchar_t*>(display_name));

		malloc_interface->Free(display_name);

		name_list.push_back(text);

		moniker_interface->Release();
	}

	malloc_interface->Release();
	moniker_enum_interface->Release();
	running_object_interface->Release();
	bind_context_interface->Release();

#endif

	return name_list;
}

TAutomation* TAutomation::GetSubObject(
    const QString 						&name)
{
#ifdef Q_OS_WIN

	VARIANT								return_variant;
	HRESULT								result_status;
	DISPID								dispatch_identification;
	DISPPARAMS							dispatch_parameters;
	BSTR 								parameter_name;

	if(!d_dispatch_interface)
	{
		return 0;
	}

	if(!name.length())
	{
		return 0;
	}

	::VariantInit(&return_variant);

	parameter_name = ::SysAllocString(reinterpret_cast<const wchar_t*>(name.utf16()));

	d_dispatch_interface->GetIDsOfNames(
				    IID_NULL,
				    &parameter_name,
				    1,
				    LOCALE_SYSTEM_DEFAULT,
				    &dispatch_identification);

	::SysFreeString(parameter_name);

	if (dispatch_identification == -1)
	{
		d_last_error = QStringLiteral("GetSubObject Error: ") + name;
		return 0;
	}

	dispatch_parameters.rgvarg = 0;
	dispatch_parameters.rgdispidNamedArgs = 0;
	dispatch_parameters.cArgs = 0;
	dispatch_parameters.cNamedArgs = 0;

	result_status = d_dispatch_interface->Invoke(
	             dispatch_identification,
	             IID_NULL,
	             LOCALE_SYSTEM_DEFAULT,
	             DISPATCH_PROPERTYGET,
	             &dispatch_parameters,
	             &return_variant,
	             0,
	             0);

	if(HRESULT_CODE(result_status) != S_OK)
	{
		d_last_error = QStringLiteral("GetSubObject Error: Object '%1' does not exist.").arg(name);
	}
	else
	{
		if(return_variant.vt == VT_BSTR)
		{
			::SysFreeString(return_variant.bstrVal);
		}

		if (return_variant.vt == VT_DISPATCH)
		{
			return new TAutomation(return_variant.pdispVal);
		}
	}
#else
	Q_UNUSED(name);

#endif

	return 0;
}

TAutomation::TArg TAutomation::GetProperty(
    const QString 						&name,
    const TAutomation::TArg 			&param_1,
    const TAutomation::TArg 			&param_2,
    const TAutomation::TArg 			&param_3,
    const TAutomation::TArg 			&param_4,
    const TAutomation::TArg 			&param_5,
    const TAutomation::TArg 			&param_6,
    const TAutomation::TArg 			&param_7,
    const TAutomation::TArg 			&param_8,
    const TAutomation::TArg 			&param_9,
    const TAutomation::TArg 			&param_10,
    const TAutomation::TArg 			&param_11,
    const TAutomation::TArg 			&param_12,
    const TAutomation::TArg 			&param_13,
    const TAutomation::TArg 			&param_14,
    const TAutomation::TArg 			&param_15,
    const TAutomation::TArg 			&param_16)
{
#ifdef Q_OS_WIN

	VARIANT								return_variant;
	HRESULT								result_status;
	DISPID								dispatch_identification;
	DISPPARAMS							dispatch_parameters;
	VARIANTARG							variant_argument[16];
	int									variant_argument_count;
	int									cntr;
	BSTR								parameter_name;

	if (d_dispatch_interface == 0)
	{
		d_last_error = QString("Call Error: d_dispatch_interface NULL");
		return TAutomation::TArg();
	}

	if(!name.length())
	{
		d_last_error = QString("Call Error: Method name length is zero.");
		return TAutomation::TArg();
	}

	VariantInit(&return_variant);

	parameter_name = ::SysAllocString(reinterpret_cast<const wchar_t*>(name.utf16()));

	d_dispatch_interface->GetIDsOfNames(
				    IID_NULL,
				    &parameter_name,
				    1,
				    LOCALE_SYSTEM_DEFAULT,
				    &dispatch_identification);

	::SysFreeString(parameter_name);

	if (dispatch_identification == -1)
	{
		d_last_error = QString("Call Error: Method '%1' does not exist.").arg(name);

#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::GetProperty()  dispatch_identification fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif

		return TAutomation::TArg();
	}

	variant_argument_count = 0;

	if(!param_16.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_16))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}

	}

	if(!param_15.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_15))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_14.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_14))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_13.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_13))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_12.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_12))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_11.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_11))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_10.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_10))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_9.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_9))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_8.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_8))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_7.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_7))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_6.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_6))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_5.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_5))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_4.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_4))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_3.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_3))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_2.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_2))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_1.IsEmpty())
	{
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_1))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	dispatch_parameters.rgvarg = &variant_argument[0];
	dispatch_parameters.rgdispidNamedArgs = 0;
	dispatch_parameters.cArgs = variant_argument_count;
	dispatch_parameters.cNamedArgs = 0;

	result_status = d_dispatch_interface->Invoke(
	             dispatch_identification,
	             IID_NULL,
	             LOCALE_SYSTEM_DEFAULT,
	             DISPATCH_PROPERTYGET,
	             &dispatch_parameters,
	             &return_variant,
	             0,
	             0);

	for (cntr=0;cntr<variant_argument_count;cntr++)
	{
		if(variant_argument[cntr].vt == VT_BSTR)
		{
			::SysFreeString(variant_argument[cntr].bstrVal);
		}

		if(variant_argument[cntr].vt == (VT_VARIANT | VT_ARRAY))
		{
			::SafeArrayDestroy(variant_argument[cntr].parray);
		}
	}

	if(HRESULT_CODE(result_status) != S_OK)
	{
		d_last_error = QStringLiteral("GetProperly Error: ") + Format_Status_Text(result_status);
		
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::GetProperty()  Invoke fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif
	}
	else
	{
		if(return_variant.vt == VT_BOOL)
		{
			return TAutomation::TArg(return_variant.iVal != 0);
		}
		else if(return_variant.vt == VT_I1)
		{
			return TAutomation::TArg(static_cast<int>(return_variant.bVal));
		}
		else if(return_variant.vt == VT_I4)
		{
			return TAutomation::TArg(static_cast<int>(return_variant.lVal));
		}
		else if(return_variant.vt == VT_BSTR)
		{
			TAutomation::TArg return_string(
								QString::fromWCharArray(
											return_variant.bstrVal,
											lstrlen(return_variant.bstrVal)));

			::SysFreeString(return_variant.bstrVal);

			return return_string;
		}
		else if(return_variant.vt == VT_R4)
		{
			return TAutomation::TArg(return_variant.fltVal);
		}
		else if(return_variant.vt == VT_R8)
		{
			return TAutomation::TArg(return_variant.dblVal);
		}
		else if (return_variant.vt == VT_DISPATCH)
		{
			return TAutomation::TArg(return_variant.pdispVal);
		}
	}
#else
	Q_UNUSED(name);
	Q_UNUSED(param_1);
	Q_UNUSED(param_2);
	Q_UNUSED(param_3);
	Q_UNUSED(param_4);
	Q_UNUSED(param_5);
	Q_UNUSED(param_6);
	Q_UNUSED(param_7);
	Q_UNUSED(param_8);
	Q_UNUSED(param_9);
	Q_UNUSED(param_10);
	Q_UNUSED(param_11);
	Q_UNUSED(param_12);
	Q_UNUSED(param_13);
	Q_UNUSED(param_14);
	Q_UNUSED(param_15);
	Q_UNUSED(param_16);

#endif

	return TAutomation::TArg();
}

bool TAutomation::CreateObject(
    const QString						&name)
{
#ifdef Q_OS_WIN
	HRESULT 							result_status;
	CLSID								clsid_application;

 	this->ReleaseObject();

	if(name.size() == 0)
	{
		d_last_error = QString("ERR  CreateObect: Object '%1' name size is zero.").arg(name);
		return false;
	}

	if(!name.length())
	{
		d_last_error = QStringLiteral("ERR  CreateObect: name of object is missing.");
		return false;
	}

	if(S_OK != CLSIDFromProgID(
	            reinterpret_cast<const wchar_t*>(name.utf16()),
	            &clsid_application))
	{
		d_last_error = QString("ERR  CreateObect: Object '%1' does not exist.").arg(name);
		return false;
	}

	result_status = CoCreateInstance(
	             clsid_application,
	             0,
	             CLSCTX_SERVER,
	             IID_IUnknown,
	             reinterpret_cast<void**>(&d_unknown_interface));

	if(d_unknown_interface != 0 && SUCCEEDED(result_status) != 0)
	{
		result_status = d_unknown_interface->QueryInterface(
		             IID_IDispatch,
		             reinterpret_cast<void**>(&d_dispatch_interface));

		if(SUCCEEDED(result_status))
		{
			return true;
		}
		else
		{
			d_last_error = QStringLiteral("ERR  QueryInterface: ") + this->Format_Status_Text(result_status);
		}
	}
	else
	{
		if(static_cast<unsigned int>(result_status) == 0x800702e4)
		{
			d_last_error = QStringLiteral("ERR  CoCreateInstance: result_status 0x800702e4 - The Requested Operation Requires Elevation.");
		}
		else
		{
			d_last_error = QString("ERR  QueryInterface: result_status 0x%1 - ").arg(result_status,0,16,QChar('0')) + this->Format_Status_Text(result_status);
		}
	}
#else
	Q_UNUSED(name);

#endif

	return false;
}

bool TAutomation::CreateObjectFromCLSID(
	unsigned char * const				CLSID_bytes)
{
#ifdef Q_OS_WIN
	HRESULT 							result_status;
	CLSID								clsid_application;
	unsigned char						*CLSID_data(reinterpret_cast<unsigned char*>(&clsid_application));
	
	this->ReleaseObject();
	
	if(CLSID_bytes == 0)
	{
		d_last_error = QStringLiteral("ERR  CreateObect: Object CLSID is null.");
		return false;
	}
	
	CLSID_data[0] = CLSID_bytes[3];
	CLSID_data[1] = CLSID_bytes[2];
	CLSID_data[2] = CLSID_bytes[1];
	CLSID_data[3] = CLSID_bytes[0];
	CLSID_data[4] = CLSID_bytes[5];
	CLSID_data[5] = CLSID_bytes[4];
	CLSID_data[6] = CLSID_bytes[7];
	CLSID_data[7] = CLSID_bytes[6];
	CLSID_data[8] = CLSID_bytes[8];
	CLSID_data[9] = CLSID_bytes[9];
	CLSID_data[10] = CLSID_bytes[10];
	CLSID_data[11] = CLSID_bytes[11];
	CLSID_data[12] = CLSID_bytes[12];
	CLSID_data[13] = CLSID_bytes[13];
	CLSID_data[14] = CLSID_bytes[14];
	CLSID_data[15] = CLSID_bytes[15];
	
		
	result_status = CoCreateInstance(
									 clsid_application,
									 0,
									 CLSCTX_SERVER,
									 IID_IUnknown,
									 reinterpret_cast<void**>(&d_unknown_interface));
	
	if(d_unknown_interface != 0 && SUCCEEDED(result_status) != 0)
	{
		result_status = d_unknown_interface->QueryInterface(
		             IID_IDispatch,
		             reinterpret_cast<void**>(&d_dispatch_interface));
		
		if(SUCCEEDED(result_status))
		{
			return true;
		}
		else
		{
			d_last_error = QStringLiteral("ERR  QueryInterface: ") + this->Format_Status_Text(result_status);
		}
	}
	else
	{
		if(static_cast<unsigned int>(result_status) == 0x800702e4)
		{
			d_last_error = QStringLiteral("ERR  CoCreateInstance: result_status 0x800702e4 - The Requested Operation Requires Elevation.");
		}
		else
		{
			d_last_error = QString("ERR  QueryInterface: result_status 0x%1 - ").arg(result_status,0,16,QChar('0')) + this->Format_Status_Text(result_status);
		}
	}
#else
	Q_UNUSED(CLSID_bytes);

#endif
	
	return false;
}


void TAutomation::ReleaseObject(void)
{
	this->Disable_Events();

#ifdef Q_OS_WIN
	
	if(d_dispatch_interface)
	{
		d_dispatch_interface->Release();
		d_dispatch_interface = 0;
	}

	if(d_unknown_interface)
	{
		d_unknown_interface->Release();
		d_unknown_interface = 0;
	}
	
#endif
}

bool TAutomation::ConnectToRunningObject(
    const QString						&name)
{
#ifdef Q_OS_WIN

	IBindCtx                            *bind_context_interface;
	IRunningObjectTable                 *running_object_interface;
	IEnumMoniker                        *moniker_enum_interface;
	IMalloc                             *malloc_interface;
	ULONG                               actual_retrieved;
	IMoniker                            *moniker_interface;
	wchar_t                             *display_name(0);
	HRESULT                             result_status;
	QString								text;

	this->ReleaseObject();

	result_status = CoGetMalloc(1,&malloc_interface);

	if(FAILED(result_status))
	{
		d_last_error = QStringLiteral("ConnectToRunningObject Error: ") + this->Format_Status_Text(result_status);
		return false;
	}

	result_status = CreateBindCtx(0,&bind_context_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();

		d_last_error = QStringLiteral("ConnectToRunningObject Error: ") + this->Format_Status_Text(result_status);
		return false;
	}

	result_status = bind_context_interface->GetRunningObjectTable(&running_object_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();
		bind_context_interface->Release();

		d_last_error = QStringLiteral("ConnectToRunningObject Error: ") + this->Format_Status_Text(result_status);
		return false;
	}

	result_status = running_object_interface->EnumRunning(&moniker_enum_interface);

	if(FAILED(result_status))
	{
		malloc_interface->Release();
		running_object_interface->Release();
		bind_context_interface->Release();

		d_last_error = QStringLiteral("ConnectToRunningObject Error: ") + this->Format_Status_Text(result_status);
		return false;
	}

	moniker_enum_interface->Reset();

	while(moniker_enum_interface->Next(1,&moniker_interface,&actual_retrieved) == S_OK)
	{
		moniker_interface->GetDisplayName(bind_context_interface,0,&display_name);

		text = QString::fromWCharArray(static_cast<wchar_t*>(display_name));

		malloc_interface->Free(display_name);

		if(0 == QString::compare(name,text,Qt::CaseInsensitive))
		{
			result_status=moniker_interface->BindToObject(
			           bind_context_interface,
			           0,
			           IID_IDispatch,
			           reinterpret_cast<void**>(&d_dispatch_interface));

			if(SUCCEEDED(result_status))
			{
				malloc_interface->Release();
				moniker_interface->Release();
				moniker_enum_interface->Release();
				running_object_interface->Release();
				bind_context_interface->Release();

				return true;
			}

		}
		moniker_interface->Release();
	}

	malloc_interface->Release();
	moniker_enum_interface->Release();
	running_object_interface->Release();
	bind_context_interface->Release();

#else
	Q_UNUSED(name);
#endif

	return false;
}

bool TAutomation::SetProperty(
    const QString 						&name,
    const TAutomation::TArg 			&param)
{

#ifdef Q_OS_WIN

	VARIANT								return_variant;
	HRESULT								result_status;
	DISPID								dispatch_identification;
	DISPID								dispidsNamedArgs(DISPID_PROPERTYPUT);
	DISPPARAMS							dispatch_parameters;
	VARIANTARG							variant_argument;
	BSTR								parameter_name;

	if (d_dispatch_interface == 0)
	{
		d_last_error = QString("Call Error: d_dispatch_interface NULL");
		return false;
	}

	if(!name.length())
	{
		d_last_error = QString("Call Error: Method name length is zero.");
		return false;
	}

	if(param.GetType() == TAutomation::TArg::ARG_EMPTY)
	{
		d_last_error = QString("SetProperty Error: Property '%1' cannot be empty.").arg(name);
		return false;
	}

	VariantInit(&return_variant);
	VariantInit(&variant_argument);

	parameter_name = ::SysAllocString(reinterpret_cast<const wchar_t*>(name.utf16()));

	d_dispatch_interface->GetIDsOfNames(
				    IID_NULL,
				    &parameter_name,
				    1,
				    LOCALE_SYSTEM_DEFAULT,
				    &dispatch_identification);

	::SysFreeString(parameter_name);

	if (dispatch_identification == -1)
	{
		d_last_error = QString("SetProperty Error: Property '%1' does not exist.").arg(name);
		
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::SetProperty()  dispatch_identification fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif
		return false;
	}

	if(!this->Pack_Argument(&variant_argument,param))
	{
		return false;
	}


	dispatch_parameters.rgvarg = &variant_argument;
	dispatch_parameters.rgdispidNamedArgs = &dispidsNamedArgs;
	dispatch_parameters.cArgs = 1;
	dispatch_parameters.cNamedArgs = 1;

	result_status = d_dispatch_interface->Invoke(
	             dispatch_identification,
	             IID_NULL,
	             LOCALE_SYSTEM_DEFAULT,
	             DISPATCH_PROPERTYPUT,
	             &dispatch_parameters,
	             &return_variant,
	             0,
	             0);

	if(variant_argument.vt == VT_BSTR)
	{
		::SysFreeString(variant_argument.bstrVal);
	}

	if(variant_argument.vt == (VT_VARIANT | VT_ARRAY))
	{
		::SafeArrayDestroy(variant_argument.parray);
	}

	if(HRESULT_CODE(result_status) == S_OK)
	{
		return true;
	}
	else
	{
		d_last_error = QString("SetProperty Error: ") + this->Format_Status_Text(result_status);

#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::SetProperty()  Invoke fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif

	}
	
#else

	Q_UNUSED(name);
	Q_UNUSED(param);

#endif

	return false;
}


TAutomation::TArg TAutomation::Call(
    const QString 						&name,
    const TAutomation::TArg 			&param_1,
    const TAutomation::TArg 			&param_2,
    const TAutomation::TArg 			&param_3,
    const TAutomation::TArg 			&param_4,
    const TAutomation::TArg 			&param_5,
    const TAutomation::TArg 			&param_6,
    const TAutomation::TArg 			&param_7,
    const TAutomation::TArg 			&param_8,
    const TAutomation::TArg 			&param_9,
    const TAutomation::TArg 			&param_10,
    const TAutomation::TArg 			&param_11,
    const TAutomation::TArg 			&param_12,
    const TAutomation::TArg 			&param_13,
    const TAutomation::TArg 			&param_14,
    const TAutomation::TArg 			&param_15,
    const TAutomation::TArg 			&param_16)
{
#ifdef Q_OS_WIN

	VARIANT								return_variant;
	HRESULT								result_status;
	DISPID								dispatch_identification;
	DISPPARAMS							dispatch_parameters;
	VARIANTARG							variant_argument[16];
	int									variant_argument_count;
	int									cntr;
	BSTR								parameter_name;


	if (d_dispatch_interface == 0)
	{

		d_last_error = QString("Call Error: d_dispatch_interface NULL");
		return TAutomation::TArg();
	}

	if(!name.length())
	{
		d_last_error = QString("Call Error: Method name length is zero.");
		return TAutomation::TArg();
	}

	VariantInit(&return_variant);

#ifdef DEBUG_AUTOMATION
	qWarning(QString("TAutomation::Call().  Parameter Name: %1").arg(name).toLatin1());
#endif

	parameter_name = ::SysAllocString(reinterpret_cast<const wchar_t*>(name.utf16()));

	d_dispatch_interface->GetIDsOfNames(
				    IID_NULL,
				    &parameter_name,
				    1,
				    LOCALE_SYSTEM_DEFAULT,
				    &dispatch_identification);

	::SysFreeString(parameter_name);
	


	if (dispatch_identification == -1)
	{
		d_last_error = QString("Call: Method '%1' does not exist.").arg(name);
		
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call()  dispatch_identification fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif

		return TAutomation::TArg();
	}

	variant_argument_count = 0;

	if(!param_16.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 16: %1").arg(param_16.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_16))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}

	}

	if(!param_15.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 15: %1").arg(param_15.AsString()).toLatin1());
#endif
		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_15))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_14.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 14: %1").arg(param_14.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_14))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_13.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 13: %1").arg(param_13.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_13))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_12.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 12: %1").arg(param_12.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_12))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_11.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 11: %1").arg(param_11.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_11))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_10.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 10: %1").arg(param_10.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_10))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_9.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 9: %1").arg(param_9.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_9))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_8.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 8: %1").arg(param_8.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_8))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_7.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 7: %1").arg(param_7.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_7))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_6.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 6: %1").arg(param_6.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_6))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_5.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 5: %1").arg(param_5.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_5))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_4.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 4: %1").arg(param_4.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_4))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_3.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 3: %1").arg(param_3.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_3))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_2.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 2: %1").arg(param_2.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_2))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	if(!param_1.IsEmpty())
	{
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call param 1: %1").arg(param_1.AsString()).toLatin1());
#endif

		if(this->Pack_Argument(&variant_argument[variant_argument_count],param_1))
		{
			variant_argument_count++;
		}
		else
		{
			return TAutomation::TArg();
		}
	}

	dispatch_parameters.rgvarg = &variant_argument[0];
	dispatch_parameters.rgdispidNamedArgs = 0;
	dispatch_parameters.cArgs = variant_argument_count;
	dispatch_parameters.cNamedArgs = 0;

	result_status = d_dispatch_interface->Invoke(
	             dispatch_identification,
	             IID_NULL,
	             LOCALE_SYSTEM_DEFAULT,
	             DISPATCH_METHOD,
		         &dispatch_parameters,
		         &return_variant,
		         0,
		         0);

#ifdef DEBUG_AUTOMATION
	qWarning(QString("TAutomation::Call return_variant.vt: %1").arg(return_variant.vt).toLatin1());
#endif

	for (cntr=0;cntr<variant_argument_count;cntr++)
	{
		if(variant_argument[cntr].vt == VT_BSTR)
		{
			::SysFreeString(variant_argument[cntr].bstrVal);
		}

		if(variant_argument[cntr].vt == (VT_VARIANT | VT_ARRAY))
		{
			::SafeArrayDestroy(variant_argument[cntr].parray);
		}
	}

	if(HRESULT_CODE(result_status) != S_OK)
	{
		d_last_error = QStringLiteral("Call Error: ") + this->Format_Status_Text(result_status);
		
#ifdef DEBUG_AUTOMATION
		qWarning(QString("TAutomation::Call()  Invoke fail.  Last error: %1").arg(d_last_error).toLatin1());
#endif

	}
	else
	{
		if(return_variant.vt == VT_BOOL)
		{
			return TAutomation::TArg(return_variant.iVal != 0);
		}
		else if(return_variant.vt == VT_I1)
		{
			return TAutomation::TArg(static_cast<int>(return_variant.bVal));
		}
		else if(return_variant.vt == VT_I4)
		{
			return TAutomation::TArg(static_cast<int>(return_variant.lVal));
		}
		else if(return_variant.vt == VT_BSTR)
		{
			TAutomation::TArg RetString(
								QString::fromWCharArray(
											return_variant.bstrVal,
											lstrlen(return_variant.bstrVal)));

			::SysFreeString(return_variant.bstrVal);

			return RetString;
		}
		else if(return_variant.vt == VT_R4)
		{
			return TAutomation::TArg(return_variant.fltVal);
		}
		else if(return_variant.vt == VT_R8)
		{
			return TAutomation::TArg(return_variant.dblVal);
		}
		else if (return_variant.vt == VT_DISPATCH)
		{
			return TAutomation::TArg(return_variant.pdispVal);
		}
	}
#else
	Q_UNUSED(name);
	Q_UNUSED(param_1);
	Q_UNUSED(param_2);
	Q_UNUSED(param_3);
	Q_UNUSED(param_4);
	Q_UNUSED(param_5);
	Q_UNUSED(param_6);
	Q_UNUSED(param_7);
	Q_UNUSED(param_8);
	Q_UNUSED(param_9);
	Q_UNUSED(param_10);
	Q_UNUSED(param_11);
	Q_UNUSED(param_12);
	Q_UNUSED(param_13);
	Q_UNUSED(param_14);
	Q_UNUSED(param_15);
	Q_UNUSED(param_16);

#endif

	return TAutomation::TArg();
}

bool TAutomation::Init_Events(
	const QString 						&type_library,
	const QString 						&class_name)
{
#ifdef Q_OS_WIN

	HRESULT								result_status;
	bool								return_status(false);
	unsigned int 						count;
	unsigned int						cntr;
	BSTR								name;
   	QString								text;
	ITypeLib 							*type_library_interface(0);
	ITypeInfo 							*type_information_interface(0);
    TYPEATTR 							*type_attribute_struct(0);

	result_status = LoadTypeLib(reinterpret_cast<const wchar_t*>(QDir::toNativeSeparators(type_library).utf16()), &type_library_interface);
	
	if(result_status != S_OK)
	{
		d_last_error = QString("Init_Events: Cannot load Type Library '%1'").arg(type_library);
		return false;
	}
		
	count = type_library_interface->GetTypeInfoCount();

	for(cntr = 0;cntr < count;++cntr)
	{
		if(S_OK == type_library_interface->GetTypeInfo(cntr,&type_information_interface))
		{   					
			type_information_interface->GetDocumentation(MEMBERID_NIL, &name, 0, 0, 0);
			text = QString::fromWCharArray(name);
			::SysFreeString(name);
			
			if(text == class_name)
			{	
				if(S_OK == type_information_interface->GetTypeAttr(&type_attribute_struct))
				{
					memcpy(&d_automation_event_iid,&type_attribute_struct->guid,sizeof(IID));
					type_information_interface->ReleaseTypeAttr(type_attribute_struct);
					
					type_information_interface->Release();		
					return_status = true;
					
					break;
				}
			}
			
			type_information_interface->Release();		
		}
	}
		
	type_library_interface->Release();

	return return_status;
#else
	
	Q_UNUSED(type_library);
	Q_UNUSED(class_name);
	
	return false;
#endif
}	
	
bool TAutomation::Init_Event_Signals(
	const QString 						&type_library)
{
#ifdef Q_OS_WIN
	
	HRESULT								result_status;
	bool								return_status(false);
	unsigned int 						count;
	unsigned int						cntr;
	const unsigned int					NAME_ARRAY_MAXCOUNT(255);
	BSTR 								name_array[NAME_ARRAY_MAXCOUNT];
	unsigned int						name_array_cntr;
   	unsigned int 						name_array_count;
   	unsigned int						member_cntr;
   	QString								text;
   	TAutomation::TSinkSignal			sink_signal;
	ITypeLib 							*type_library_interface(0);
	ITypeInfo 							*type_information_interface(0);
    TYPEATTR 							*type_attribute_struct(0);
	FUNCDESC 							*function_description_struct(0);

	d_event_signals.clear();
	d_signal_list.clear();

	result_status = ::LoadTypeLib(reinterpret_cast<const wchar_t*>(QDir::toNativeSeparators(type_library).utf16()), &type_library_interface);
	
	if(result_status != S_OK)
	{
		d_last_error = QString("Init_Event_Signals: Cannot load Type Library '%1'").arg(type_library);
		return false;
	}
		
	count = type_library_interface->GetTypeInfoCount();

	for(cntr = 0;cntr < count;++cntr)
	{
		if(S_OK == type_library_interface->GetTypeInfo(cntr,&type_information_interface))
		{   	
			if(S_OK == type_information_interface->GetTypeAttr(&type_attribute_struct))
			{
				if(0 == memcmp(&type_attribute_struct->guid,&d_automation_event_iid,sizeof(IID)))
				{
					return_status = true;

					// Get all values of the enumeration
					for (member_cntr = 0; member_cntr < static_cast<unsigned int>(type_attribute_struct->cFuncs); ++member_cntr) 
					{
						type_information_interface->GetFuncDesc(member_cntr, &function_description_struct);
						type_information_interface->GetNames(function_description_struct->memid, (BSTR*)&name_array, NAME_ARRAY_MAXCOUNT, &name_array_count);
						sink_signal.id_index = static_cast<int>(function_description_struct->memid);
						
						for (name_array_cntr = 0; name_array_cntr < name_array_count; ++name_array_cntr) 
						{
							text = QString::fromWCharArray(name_array[name_array_cntr]);
							
							if(name_array_cntr == 0)
							{	
								sink_signal.id_name = text;
							}
							else
							{
								sink_signal.parameters.push_back(text);
							}
							
							::SysFreeString(name_array[name_array_cntr]);
						}

						type_information_interface->ReleaseFuncDesc(function_description_struct);
						
						d_event_signals.push_back(sink_signal);
					}
				}
				type_information_interface->ReleaseTypeAttr(type_attribute_struct);
			}
			type_information_interface->Release();		
		}
		
		if(return_status)
		{
			break;
		}
	}
		
	type_library_interface->Release();

	return return_status;
#else

	Q_UNUSED(type_library);	
	return false;
	
#endif
}

bool TAutomation::Add_Event_Signal(
	const QString 						&name)
{
#ifdef Q_OS_WIN

	std::vector<TAutomation::TSinkSignal>::const_iterator iter;
		
	for(iter = d_event_signals.begin();iter != d_event_signals.end();++iter)
	{
		if((*iter).id_name == name)
		{
			d_signal_list[(*iter).id_index] = (*iter).id_name;			
			return true;
		}
	}
#else

	Q_UNUSED(name);
	
#endif
	
	return false;
}

bool TAutomation::Enable_Events(void)
{
#ifdef Q_OS_WIN

	IConnectionPointContainer 			*connection_point_container(0);
	HRESULT 							result_status;
		
	if(d_event_signals.size() == 0)
	{
		return false;
	}
	
	if(!d_dispatch_interface)
	{
		return false;
	}
	
	if(d_automation_event_connection_point && d_automation_event_connection_id > 0)
	{
	    d_automation_event_connection_point->Unadvise(d_automation_event_connection_id);
        d_automation_event_connection_point->Release();
        d_automation_event_connection_point = 0;
        d_automation_event_connection_id = 0;
        
        // this is automatically deleted from 'Unadvise'
        d_automation_event= 0;
	}
	else if(d_automation_event)
	{
		d_automation_event->Release();		
		d_automation_event= 0;
	}
	
	d_dispatch_interface->QueryInterface(IID_IConnectionPointContainer, (void**)&connection_point_container);
		
	if(connection_point_container)
	{
		result_status = connection_point_container->FindConnectionPoint(d_automation_event_iid,&d_automation_event_connection_point);
	
		if(result_status != S_OK)
		{
			connection_point_container->Release();
			return false;
		}	
	
		d_automation_event = new TAutomationEvent(this,d_automation_event_iid);
	
		d_automation_event_connection_point->Advise(reinterpret_cast<IUnknown*>(d_automation_event),&d_automation_event_connection_id);
		
		if(!(d_automation_event_connection_id > 0))	// valid connection points are always greater than 0
		{
			d_automation_event->Release();
			d_automation_event = 0;
			
			d_automation_event_connection_point->Release();
			connection_point_container->Release();

			return false;
		}
		
		connection_point_container->Release();	
		return true;
	}
	
		
#endif

	return false;
}

void TAutomation::Disable_Events(void)
{
#ifdef Q_OS_WIN

	d_signal_list.clear();
	d_event_signals.clear();

	if(d_automation_event_connection_point && d_automation_event_connection_id > 0)
	{
		d_automation_event_connection_point->Unadvise(d_automation_event_connection_id);
		d_automation_event_connection_point->Release();
		d_automation_event_connection_point = 0;
		d_automation_event_connection_id = 0;

		// this is automatically deleted from 'Unadvise'
		d_automation_event = 0;
	}
	else if(d_automation_event)
	{
		d_automation_event->Release();
		d_automation_event = 0;
	}

#endif
}

bool TAutomation::Test_For_Function(
	const QString 						&type_library,
	const QString 						&interface_name,
	const QString 						&function_name)
{
#ifdef Q_OS_WIN
	
	HRESULT								result_status;
	bool								return_status(false);
	unsigned int 						count;
	unsigned int						cntr;
	const unsigned int					NAME_ARRAY_MAXCOUNT(255);
	BSTR 								name_array[NAME_ARRAY_MAXCOUNT];
	BSTR								name;
	unsigned int						name_array_cntr;
   	unsigned int 						name_array_count;
   	unsigned int						member_cntr;
   	QString								text;
	ITypeLib 							*type_library_interface(0);
	ITypeInfo 							*type_information_interface(0);
    TYPEATTR 							*type_attribute_struct(0);
	FUNCDESC 							*function_description_struct(0);


	result_status = LoadTypeLib(reinterpret_cast<const wchar_t*>(QDir::toNativeSeparators(type_library).utf16()), &type_library_interface);
	
	if(result_status != S_OK)
	{
		return false;
	}
		
	count = type_library_interface->GetTypeInfoCount();

	for(cntr = 0;cntr < count;++cntr)
	{
		if(S_OK == type_library_interface->GetTypeInfo(cntr,&type_information_interface))
		{   	
			type_information_interface->GetDocumentation(MEMBERID_NIL, &name, 0, 0, 0);
			text = QString::fromWCharArray(name);
			::SysFreeString(name);
			
			if(text == interface_name)
			{				
				if(S_OK == type_information_interface->GetTypeAttr(&type_attribute_struct))
				{
					// Get all values of the enumeration
					for (member_cntr = 0; member_cntr < static_cast<unsigned int>(type_attribute_struct->cFuncs); ++member_cntr) 
					{
						type_information_interface->GetFuncDesc(member_cntr, &function_description_struct);				
						type_information_interface->GetNames(function_description_struct->memid, (BSTR*)&name_array, NAME_ARRAY_MAXCOUNT, &name_array_count);
					
						for (name_array_cntr = 0; name_array_cntr < name_array_count; ++name_array_cntr) 
						{
							text = QString::fromWCharArray(name_array[name_array_cntr]);						
							::SysFreeString(name_array[name_array_cntr]);
							
							if(text == function_name)
							{
								return_status = true;
							}
						}

						type_information_interface->ReleaseFuncDesc(function_description_struct);
						
						if(return_status)
						{
							break;
						}
					}
					
					type_information_interface->ReleaseTypeAttr(type_attribute_struct);
				}
			}
			
			type_information_interface->Release();		
		}
		
		if(return_status)
		{
			break;
		}
	}
		
	type_library_interface->Release();

	return return_status;
	
#else
	Q_UNUSED(type_library);
	Q_UNUSED(interface_name);
	Q_UNUSED(function_name);
	
	return false;
#endif
}
	
bool TAutomation::Test_For_Property(
	const QString 						&type_library,
	const QString 						&interface_name,
	const QString 						&property_name)
{
#ifdef Q_OS_WIN
	
	HRESULT								result_status;
	bool								return_status(false);
	unsigned int 						count;
	unsigned int						cntr;
	const unsigned int					NAME_ARRAY_MAXCOUNT(255);
	BSTR 								name_array[NAME_ARRAY_MAXCOUNT];
	BSTR								name;
	unsigned int						name_array_cntr;
   	unsigned int 						name_array_count;
   	unsigned int						member_cntr;
   	QString								text;
	ITypeLib 							*type_library_interface(0);
	ITypeInfo 							*type_information_interface(0);
    TYPEATTR 							*type_attribute_struct(0);
	VARDESC 							*variable_description_struct(0);

	result_status = LoadTypeLib(reinterpret_cast<const wchar_t*>(QDir::toNativeSeparators(type_library).utf16()), &type_library_interface);
	
	if(result_status != S_OK)
	{
		return false;
	}
		
	count = type_library_interface->GetTypeInfoCount();

	for(cntr = 0;cntr < count;++cntr)
	{
		if(S_OK == type_library_interface->GetTypeInfo(cntr,&type_information_interface))
		{   	
			type_information_interface->GetDocumentation(MEMBERID_NIL, &name, 0, 0, 0);
			text = QString::fromWCharArray(name);
			::SysFreeString(name);
			
			if(text == interface_name)
			{				
				if(S_OK == type_information_interface->GetTypeAttr(&type_attribute_struct))
				{
					// Get all values of the enumeration
					for (member_cntr = 0; member_cntr < static_cast<unsigned int>(type_attribute_struct->cVars); ++member_cntr) 
					{
						type_information_interface->GetVarDesc(member_cntr, &variable_description_struct);				
						type_information_interface->GetNames(variable_description_struct->memid, (BSTR*)&name_array, NAME_ARRAY_MAXCOUNT, &name_array_count);
					
						for (name_array_cntr = 0; name_array_cntr < name_array_count; ++name_array_cntr) 
						{
							text = QString::fromWCharArray(name_array[name_array_cntr]);						
							::SysFreeString(name_array[name_array_cntr]);
							
							if(text == property_name)
							{
								return_status = true;
							}
						}

						type_information_interface->ReleaseVarDesc(variable_description_struct);
						
						if(return_status)
						{
							break;
						}
					}
					
					type_information_interface->ReleaseTypeAttr(type_attribute_struct);
				}
			}
			
			type_information_interface->Release();		
		}
		
		if(return_status)
		{
			break;
		}
	}
		
	type_library_interface->Release();

	return return_status;
	
#else

	Q_UNUSED(type_library);
	Q_UNUSED(interface_name);
	Q_UNUSED(property_name);
	
	return false;
	
#endif
}

bool TAutomation::InitializeComLibrary(void)
{
#ifdef Q_OS_WIN

	return (S_OK == ::CoInitializeEx(0,COINIT_APARTMENTTHREADED));

#else
	return true;
#endif
}

bool TAutomation::InitializeComLibrary_MT(void)
{
#ifdef Q_OS_WIN

	return (S_OK == ::CoInitializeEx(0,COINIT_MULTITHREADED));

#else
	return true;
#endif
}

void TAutomation::UninitializeComLibrary(void)
{
#ifdef Q_OS_WIN

	::CoUninitialize();

#endif
}

bool TAutomation::DoesWindowClassExist(
	const QString 						&class_name)
{
#ifdef Q_OS_WIN

	if(FindWindowW(reinterpret_cast<const wchar_t*>(class_name.utf16()),0))
	{
		return true;
	}
#else
	
	Q_UNUSED(class_name);
	
#endif

	return false;

}

void TAutomation::Release(
	TAutomation 						*Aut)
{
	if(Aut)
	{
		delete Aut;
	}
}

QString TAutomation::Format_Status_Text(
	const HRESULT 						result_status)
{
	QString								text;
	
#ifdef Q_OS_WIN

    wchar_t 							*message_buffer;

    ::FormatMessageW(
				FORMAT_MESSAGE_ALLOCATE_BUFFER | 
				FORMAT_MESSAGE_FROM_SYSTEM |
				FORMAT_MESSAGE_IGNORE_INSERTS |
				FORMAT_MESSAGE_FROM_HMODULE,
				0,
				result_status,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				(LPTSTR)&message_buffer,
				0, 
				0);
    
    text = QString::fromWCharArray(message_buffer);

	::LocalFree(message_buffer);
	
#else
	Q_UNUSED(result_status);
#endif
	
	return text;
}

bool TAutomation::Pack_Argument(
    tagVARIANT 							* const variant_argument,
    const TAutomation::TArg 			&param)
{
#ifdef Q_OS_WIN
	SAFEARRAYBOUND						safe_array_bound;
	SAFEARRAY							*safe_array;
	long int							cntr;
	HRESULT								result_status;
	VARIANT								variant;

	assert(variant_argument);

	::VariantInit(variant_argument);

	switch(param.GetType())
	{
		default:
			case TAutomation::TArg::ARG_EMPTY:				// optional value
			variant_argument->vt = static_cast<unsigned short>(DISP_E_PARAMNOTFOUND);	// not sure why this is needed
			variant_argument->scode = VT_ERROR;
			break;

		case TAutomation::TArg::ARG_BOOL:
			variant_argument->vt = VT_BOOL;
			if(param.AsBool())
			{
				variant_argument->iVal = static_cast<short>(0xffff);	// true
			}
			else
			{
				variant_argument->iVal = static_cast<short>(0x0000);	// false
			}
			break;

		case TAutomation::TArg::ARG_INTEGER:
			variant_argument->vt = VT_I4;
			variant_argument->lVal = param.AsInteger();
			break;

		case TAutomation::TArg::ARG_FLOAT:
			variant_argument->vt = VT_R4;
			variant_argument->fltVal = param.AsFloat();
			break;

		case TAutomation::TArg::ARG_DOUBLE:
			variant_argument->vt = VT_R8;
			variant_argument->dblVal = param.AsDouble();
			break;

		case TAutomation::TArg::ARG_STRING:
			variant_argument->vt = VT_BSTR;

			variant_argument->bstrVal = ::SysAllocString(reinterpret_cast<const wchar_t*>(param.AsString().utf16()));

			break;

		case TAutomation::TArg::ARG_STRINGLIST:

			safe_array_bound.lLbound = 0;
			safe_array_bound.cElements = param.StringArraySize();

			safe_array = ::SafeArrayCreate(VT_VARIANT,1,&safe_array_bound);

			if(safe_array == 0)
			{
				d_last_error = "VT_ARRAY:  Cannot create SafeArray";
				return false;
			}

			for(cntr = static_cast<long int>(safe_array_bound.lLbound);
					cntr < static_cast<long int>(safe_array_bound.cElements);
					cntr++)
			{
				::VariantInit(&variant);
				variant.vt = VT_BSTR;
				variant.bstrVal = ::SysAllocString(reinterpret_cast<const wchar_t*>(param.AsString().utf16()));

				result_status = ::SafeArrayPutElement(safe_array,&cntr,(void*)&variant);

				if(FAILED(result_status))
				{
					d_last_error = "VT_ARRAY:  Cannot put element into SafeArray";
					return false;
				}
			}

			variant_argument->vt = VT_ARRAY | VT_VARIANT;
			variant_argument->parray = safe_array;

			break;

		case TAutomation::TArg::ARG_PTR_INTEGER:
			variant_argument->vt = VT_I4 | VT_BYREF;
			variant_argument->plVal = reinterpret_cast<LONG*>(param.AsPtrInteger());
			break;

		case TAutomation::TArg::ARG_PTR_DOUBLE:
			variant_argument->vt = VT_R8 | VT_BYREF;
			variant_argument->pdblVal = param.AsPtrDouble();
			break;

		case TAutomation::TArg::ARG_PTR_DISPATCH:
			variant_argument->vt = VT_DISPATCH;
			variant_argument->pdispVal = param.AsObjectInterface();
			break;
	}
#else
	Q_UNUSED(param);
	Q_UNUSED(variant_argument);

#endif

	return true;
}

bool TAutomation::Unpack_Argument(
    const tagVARIANT 					&variant_argument,
    TAutomation::TArg 					* const param)
{
#ifdef Q_OS_WIN

	assert(param);
	
	if(variant_argument.vt == VT_EMPTY ||
			variant_argument.vt == VT_NULL ||
			(variant_argument.vt == VT_ERROR && variant_argument.scode == DISP_E_PARAMNOTFOUND))
	{
		*param = TAutomation::TArg();	// empty
	}
	else if(variant_argument.vt == VT_BOOL)
	{
		if(variant_argument.iVal == static_cast<short>(0x0000))
		{
			*param = TAutomation::TArg(false);
		}
		else
		{
			*param = TAutomation::TArg(true);
		}
		
		return true;
	}
	else if(variant_argument.vt == VT_I2)
	{
		*param = TAutomation::TArg(static_cast<int>(variant_argument.iVal));		
		return true;
	}
	else if(variant_argument.vt == VT_I4)
	{
		*param = TAutomation::TArg(static_cast<int>(variant_argument.lVal));		
		return true;
	}
	else if(variant_argument.vt == VT_R4)
	{
		*param = TAutomation::TArg(static_cast<float>(variant_argument.fltVal));		
		return true;
	}
	else if(variant_argument.vt == VT_R8)
	{
		*param = TAutomation::TArg(static_cast<double>(variant_argument.dblVal));		
		return true;
	}
	else if(variant_argument.vt == VT_BSTR)
	{
		*param = TAutomation::TArg(QString::fromWCharArray(static_cast<wchar_t*>(variant_argument.bstrVal)));		
		return true;
	}
	else if(variant_argument.vt == (VT_I4 | VT_BYREF))
	{
		*param = TAutomation::TArg(reinterpret_cast<int*>(variant_argument.plVal));
		return true;
	}
	else if(variant_argument.vt == (VT_R8 | VT_BYREF))
	{
		*param = TAutomation::TArg(static_cast<double*>(variant_argument.pdblVal));		
		return true;
	}
	else if(variant_argument.vt == VT_DISPATCH)
	{
		*param = TAutomation::TArg(static_cast<IDispatch*>(variant_argument.pdispVal));		
		return true;
	}

#else
	Q_UNUSED(param);
	Q_UNUSED(variant_argument);

#endif

	return false;
}

void TAutomation::Send_Signal(
	DISPID								dispatch_member_id,
	int									argc,
	VARIANTARG							*variant_arguments)
{
#ifdef Q_OS_WIN

	std::map<DISPID,QString>::iterator 	iter;
	int									cntr;
	TArg								argument_array[16];	// maximum of 16 arguments allowed
	
	iter = d_signal_list.find(dispatch_member_id);
	
	if(iter != d_signal_list.end() && (argc > 16) == false)
	{
		for(cntr = 0;cntr < argc;++cntr)
		{
			this->Unpack_Argument((*(variant_arguments + argc - cntr - 1)),&argument_array[cntr]);	// reverse order
		}

		emit Sink_Signal((*iter).second,argc,&argument_array[0]);
	}
	
	// free anything that should be freed
	for (cntr=0; cntr < argc;++cntr)
	{
		if((variant_arguments + cntr)->vt == VT_BSTR)
		{
			::SysFreeString((variant_arguments + cntr)->bstrVal);
		}

		if((variant_arguments + cntr)->vt == (VT_VARIANT | VT_ARRAY))
		{
			::SafeArrayDestroy((variant_arguments + cntr)->parray);
		}
	}

#else
	
	Q_UNUSED(dispatch_member_id);
	Q_UNUSED(argc);
	Q_UNUSED(variant_arguments);
	
#endif
}

