/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef AUTOMATIONHEADERFILE
#define AUTOMATIONHEADERFILE

#include <QObject>
#include <map>
#include <vector>

// #define DEBUG_AUTOMATION

#ifdef Q_OS_WIN
#include <windows.h>
#include <ole2.h>
#include <ocidl.h>
#else
// these types only exist in WIN32
typedef unsigned int HRESULT;
typedef unsigned int DISPID;
typedef unsigned int VARIANTARG;
typedef unsigned int IConnectionPoint;
#endif

class IUnknown;
class IDispatch;
class TAutomationEvent;

struct tagVARIANT;

class TAutomation : public QObject
{
Q_OBJECT

public:

// The Arg class is for arguments of the Automation class and similar to variant

class TArg
{
public:

enum TType
{
	ARG_EMPTY = 0,
	ARG_BOOL,
	ARG_INTEGER,
	ARG_FLOAT,
	ARG_DOUBLE,
	ARG_STRING,
	ARG_STRINGLIST,
	ARG_PTR_DOUBLE,
	ARG_PTR_INTEGER,
	ARG_PTR_DISPATCH
};

// CREATORS
	TArg(void);
	TArg(const TArg&);
	TArg& operator=(const TArg&);
	TArg(const bool);
	TArg(const char*);
	TArg(const QString&);
	TArg(const QStringList&);
	TArg(const float&);
	TArg(const double&);
	TArg(const int);
	TArg(double*);
	TArg(int*);
	TArg(IDispatch*);

	~TArg(void);
	
// ACCESSORS
	TType GetType(void) const {return d_value_type;}
	bool IsEmpty(void) const {return (d_value_type == ARG_EMPTY);}
	
	bool AsBool(void) const;
	QString AsString(void) const;
	QString AsStringFromArray(const int Offset) const;
	int StringArraySize(void) const;
	int AsInteger(void) const;
	double AsDouble(void) const;
	float AsFloat(void) const;
	double* AsPtrDouble(void) const;
	int* AsPtrInteger(void) const;
	TAutomation* AsObject(void) const;
	IDispatch* AsObjectInterface(void) const;
	

// MANIPULATORS


private:

union TValue
{
	bool								d_boolean;
	float								d_float;
	double								d_double;
	int									d_integer;	
	double								*d_pointer_double;
	int									*d_pointer_integer;
	IDispatch							*d_dispatch_interface;
};
	
	TType								d_value_type;
	TValue								d_value;
	QString								d_text;
	QStringList							d_text_array;
};

// CREATORS
	TAutomation(const QObject *parent=0);
	TAutomation(const QString &name,const QObject *parent=0);
	TAutomation(IDispatch *dispatch,const QObject *parent=0);
	~TAutomation(void);

// ACCESSORS
	QStringList GetRunningObjectNames(void);
	IDispatch* DispatchInterface(void) const {return d_dispatch_interface;}
	TAutomation* GetSubObject(const QString &Name);
	QString GetLastErrorText(void) const {return d_last_error;}

	TArg GetProperty(const QString &Name,
			const TArg &param1 = TArg(),
			const TArg &param2 = TArg(),
			const TArg &param3 = TArg(),
			const TArg &param4 = TArg(),
			const TArg &param5 = TArg(),
			const TArg &param6 = TArg(),
			const TArg &param7 = TArg(),
			const TArg &param8 = TArg(),
			const TArg &param9 = TArg(),
			const TArg &param10 = TArg(),
			const TArg &param11 = TArg(),
			const TArg &param12 = TArg(),
			const TArg &param13 = TArg(),
			const TArg &param14 = TArg(),
			const TArg &param15 = TArg(),
			const TArg &param16 = TArg());
	bool IsNullObject(void) const {return (d_dispatch_interface == NULL);}
	
// MANIPULATORS
	bool CreateObject(const QString &Name);
	bool CreateObjectFromCLSID(unsigned char * const CLSID_bytes);	// 16 bytes in format commonly displayed in Regedit or OleView.  
	bool ConnectToRunningObject(const QString &Name);
	void ReleaseObject(void);	
			
	bool SetProperty(const QString &Name,const TArg &param);
	TArg Call(const QString &Name,
			const TArg &param1 = TArg(),
			const TArg &param2 = TArg(),
			const TArg &param3 = TArg(),
			const TArg &param4 = TArg(),
			const TArg &param5 = TArg(),
			const TArg &param6 = TArg(),
			const TArg &param7 = TArg(),
			const TArg &param8 = TArg(),
			const TArg &param9 = TArg(),
			const TArg &param10 = TArg(),
			const TArg &param11 = TArg(),
			const TArg &param12 = TArg(),
			const TArg &param13 = TArg(),
			const TArg &param14 = TArg(),
			const TArg &param15 = TArg(),
			const TArg &param16 = TArg());

	bool Init_Events(const QString &type_library,const QString &class_name);
	bool Init_Event_Signals(const QString &type_library);	
	bool Add_Event_Signal(const QString &name);
	
	bool Enable_Events(void);
	void Disable_Events(void);
	

// STATIC	
	static bool Test_For_Function(const QString &type_library,const QString &interface_name,const QString &function_name);	
	static bool Test_For_Property(const QString &type_library,const QString &interface_name,const QString &property_name);	
		
	static bool InitializeComLibrary(void);		// COINIT_APARTMENTTHREADED
	static bool InitializeComLibrary_MT(void);	// COINIT_MULTITHREADED
	static void UninitializeComLibrary(void);
	static bool DoesWindowClassExist(const QString &ClassName);
	
	// Call this to free class and release object.
	// Do not call release twice on same object
	static void Release(TAutomation *Aut);
	
signals:
	void Sink_Signal(const QString&,int,TAutomation::TArg*);
	
private:

struct TSinkSignal
{
	DISPID								id_index;
	QString								id_name;
	QStringList							parameters;
};

	IUnknown							*d_unknown_interface;
	IDispatch							*d_dispatch_interface;
	
#ifdef Q_OS_WIN
	TAutomationEvent					*d_automation_event;
	IConnectionPoint 					*d_automation_event_connection_point;
	DWORD 								d_automation_event_connection_id;
	IID									d_automation_event_iid;
	std::map<DISPID,QString>			d_signal_list;
	std::vector<TSinkSignal>			d_event_signals;
#endif

	mutable QString						d_last_error;
	
	QString Format_Status_Text(const HRESULT Status);
	bool Pack_Argument(tagVARIANT * const variant_argument, const TArg &param);
	bool Unpack_Argument(const tagVARIANT &variant_argument, TArg * const param);

	void Send_Signal(DISPID,int,VARIANTARG*);
	
// NOT IMPLEMENTED
	TAutomation(const TAutomation&);
	TAutomation& operator=(const TAutomation&);
	
	friend class TAutomationEvent;

};

// 12345678901234567890123456789012345678
//           1        2         3
// {BB208E07-C78F-11D3-91B7-00C04F796327}

#endif

